# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import itertools
from typing import List, Sequence, Set, Dict, Tuple, cast

import numpy as np
import warnings

from .PEGraphX import (
    CUNode,
    FunctionNode,
    LineID,
    LoopNode,
    NodeID,
    PEGraphX,
    NodeType,
    Node,
    DepType,
    EdgeType,
    Dependency,
    MemoryRegion,
)
from .parser import LoopData
from .variable import Variable

loop_data: Dict[LineID, int] = {}


def contains(list, filter):
    for x in list:
        if filter(x):
            return True
    return False


def correlation_coefficient(v1: List[float], v2: List[float]) -> float:
    """Calculates correlation coefficient as (A dot B) / (norm(A) * norm(B))

    :param v1: first vector
    :param v2: second vector
    :return: correlation coefficient, 0 if one of the norms is 0
    """
    norm_product = np.linalg.norm(v1) * np.linalg.norm(v2)  # type:ignore
    return 0 if norm_product == 0 else np.dot(v1, v2) / norm_product  # type:ignore


def find_subnodes(pet: PEGraphX, node: Node, criteria: EdgeType) -> List[Node]:
    """Returns direct children of a given node

    :param pet: PET graph
    :param node: CUNode
    :param criteria: EdgeType, type of edges to traverse
    :return: list of children nodes
    """
    return [pet.node_at(t) for s, t, d in pet.out_edges(node.id) if d.etype == criteria]


def depends(pet: PEGraphX, source: Node, target: Node) -> bool:
    """Detects if source node or one of it's children has a RAW dependency to target node or one of it's children

    :param pet: PET graph
    :param source: source node for dependency detection
    :param target: target of dependency
    :return: true, if there is RAW dependency
    """
    if source == target:
        return False
    target_nodes = pet.subtree_of_type(target)

    for node in pet.subtree_of_type(source, CUNode):
        for target in [
            pet.node_at(target_id)
            for source_id, target_id, dependence in pet.out_edges(node.id, EdgeType.DATA)
            if dependence.dtype == DepType.RAW
        ]:
            if target in target_nodes:
                return True
    return False


def is_loop_index2(pet: PEGraphX, root_loop: Node, var_name: str) -> bool:
    """Checks, whether the variable is a loop index.

    :param pet: CU graph
    :param root_loop: root loop
    :param var_name: name of the variable
    :return: true if variable is index of the loop
    """
    loops_start_lines = [v.start_position() for v in pet.subtree_of_type(root_loop, LoopNode)]
    return pet.is_loop_index(var_name, loops_start_lines, pet.subtree_of_type(root_loop, CUNode))


# NOTE: left old code as it may become relevant again in the near future
# We decided to omit the information that computes the workload and the relevant codes. For large programs (e.g., ffmpeg), the generated Data.xml file becomes very large. However, we keep the code here because we would like to integrate a hotspot detection algorithm (TODO: Bertin) with the parallelism discovery. Then, we need to retrieve the information to decide which code sections (loops or functions) are worth parallelizing.
def calculate_workload(pet: PEGraphX, node: Node, ignore_function_calls_and_cached_values: bool = False) -> int:
    """Calculates and stores the workload for a given node
    The workload is the number of instructions multiplied by respective number of iterations

    :param pet: PET graph
    :param node: root node
    :return: workload
    """
    # check if value already present
    if node.workload is not None:
        if not ignore_function_calls_and_cached_values:
            return node.workload
    res = 0
    if node.type == NodeType.DUMMY:
        # store workload
        node.workload = 0
        return 0
    elif node.type == NodeType.CU:
        # if a function is called, replace the instruction with the costs of the called function
        # note: recursive function calls are counted as a single instruction
        res += cast(CUNode, node).instructions_count
        if not ignore_function_calls_and_cached_values:
            for calls_edge in pet.out_edges(cast(CUNode, node).id, EdgeType.CALLSNODE):
                # add costs of the called function
                res += calculate_workload(
                    pet,
                    pet.node_at(calls_edge[1]),
                    ignore_function_calls_and_cached_values=ignore_function_calls_and_cached_values,
                )
                # substract 1 to ignore the call instruction
                # todo: should we keep the cost for the call instruction and just add the costs of the called funciton?
                res -= 1
    elif node.type == NodeType.FUNC:
        if not ignore_function_calls_and_cached_values:
            for child in find_subnodes(pet, node, EdgeType.CHILD):
                res += calculate_workload(
                    pet,
                    child,
                    ignore_function_calls_and_cached_values=ignore_function_calls_and_cached_values,
                )
    elif node.type == NodeType.LOOP:
        for child in find_subnodes(pet, node, EdgeType.CHILD):
            if child.type == NodeType.CU:
                if "for.inc" in cast(CUNode, child).basic_block_id:
                    res += cast(CUNode, child).instructions_count
                elif "for.cond" in cast(CUNode, child).basic_block_id:
                    # determine average iteration count. Use traditional iteration count as a fallback
                    average_iteration_count = (
                        cast(LoopNode, node).loop_iterations
                        if cast(LoopNode, node).loop_data is None
                        else cast(LoopData, cast(LoopNode, node).loop_data).average_iteration_count
                    )
                    res += (
                        calculate_workload(
                            pet,
                            child,
                            ignore_function_calls_and_cached_values=ignore_function_calls_and_cached_values,
                        )
                        * average_iteration_count
                        + 1
                    )
                else:
                    # determine average iteration count. Use traditional iteration count as a fallback
                    average_iteration_count = (
                        cast(LoopNode, node).loop_iterations
                        if cast(LoopNode, node).loop_data is None
                        else cast(LoopData, cast(LoopNode, node).loop_data).average_iteration_count
                    )
                    res += (
                        calculate_workload(
                            pet,
                            child,
                            ignore_function_calls_and_cached_values=ignore_function_calls_and_cached_values,
                        )
                        * average_iteration_count
                    )
            else:
                # determine average iteration count. Use traditional iteration count as a fallback
                average_iteration_count = (
                    cast(LoopNode, node).loop_iterations
                    if cast(LoopNode, node).loop_data is None
                    else cast(LoopData, cast(LoopNode, node).loop_data).average_iteration_count
                )
                res += (
                    calculate_workload(
                        pet,
                        child,
                        ignore_function_calls_and_cached_values=ignore_function_calls_and_cached_values,
                    )
                    * average_iteration_count
                )
    # store workload
    node.workload = res
    return res


def calculate_per_iteration_workload_of_loop(pet: PEGraphX, node: Node) -> int:
    """Calculates and returns the per iteration workload for a given node
    The workload is the number of instructions that happens within the loops body.
    The amount of iterations of the outer loop does not influence the result.
    :param pet: PET graph
    :param loop: root loop
    :return: workload per iteration
    """
    if node.type != NodeType.LOOP:
        raise ValueError("root has to be of type LOOP!")

    res = 0
    # sum up all children of the loops body (i.e. children excluding increment and condition)
    # in contrast to calculate_workload, do not multiply the amount of loop iterations to the workload of the children
    for child in find_subnodes(pet, node, EdgeType.CHILD):
        if child.type == NodeType.CU:
            if "for.inc" in cast(CUNode, child).basic_block_id:
                continue
            elif "for.cond" in cast(CUNode, child).basic_block_id:
                continue
            else:
                try:
                    res += calculate_workload(pet, child)
                except RecursionError as rerr:
                    warnings.warn("Cost calculation not possible for node: " + str(child.id) + "!")
        else:
            try:
                res += calculate_workload(pet, child)
            except RecursionError as rerr:
                warnings.warn("Cost calculation not possible for node: " + str(child.id) + "!")
    return res


def __get_dep_of_type(
    pet: PEGraphX, node: Node, dep_type: DepType, reversed: bool
) -> List[Tuple[NodeID, NodeID, Dependency]]:
    """Searches all dependencies of specified type

    :param pet: CU graph
    :param node: node
    :param dep_type: type of dependency
    :param reversed: if true the it looks for incoming dependencies
    :return: list of dependencies
    """
    return [
        e
        for e in (pet.in_edges(node.id, EdgeType.DATA) if reversed else pet.out_edges(node.id, EdgeType.DATA))
        if e[2].dtype == dep_type
    ]


def __get_variables(nodes: Sequence[Node]) -> Set[Variable]:
    """Gets all variables in nodes

    :param nodes: nodes
    :return: Set of variables
    """
    res = set()
    for node in nodes:
        if not isinstance(node, CUNode):
            continue
        for v in node.local_vars:
            res.add(v)
        for v in node.global_vars:
            res.add(v)
    return res


def is_reduction_var(line: LineID, name: str, reduction_vars: List[Dict[str, str]]) -> bool:
    """Determines, whether or not the given variable is reduction variable

    :param line: loop line number
    :param name: variable name
    :param reduction_vars: List of reduction variables
    :return: true if is reduction variable
    """
    return any(rv for rv in reduction_vars if rv["loop_line"] == line and rv["name"] == name)


def is_reduction_any(possible_lines: List[LineID], var_name: str, reduction_vars: List[Dict[str, str]]) -> bool:
    """Determines, whether or not the given variable is reduction variable

    :param possible_lines: possible loop line number
    :param name: variable name
    :param reduction_vars: List of reduction variables
    :return: true if is reduction variable
    """
    for line in possible_lines:
        if is_reduction_var(line, var_name, reduction_vars):
            return True

    return False


def is_written_in_subtree(
    mem_regs: Set[MemoryRegion],
    raw: Set[Tuple[NodeID, NodeID, Dependency]],
    waw: Set[Tuple[NodeID, NodeID, Dependency]],
    tree: List[Node],
) -> bool:
    """Checks if variable is written in subtree

    :param mem_reg: memory region
    :param raw: raw dependencies of the loop
    :param waw: waw dependencies of the loop
    :param tree: subtree
    :return: true if is written
    """
    for e in itertools.chain(raw, waw):
        if e[2].memory_region in mem_regs and any([n.id == e[1] for n in tree]):
            return True
    return False


def is_func_arg(pet: PEGraphX, var: str, node: Node) -> bool:
    """Checks if variable is a function argument

    :param pet: CU graph
    :param var: variable name
    :param node: loop node
    :return: true if variable is argument
    """
    # None may occur because __get_variables doesn't check for actual elements
    if var is None:
        return False
    if "." not in var:
        return False
    parents = [pet.node_at(edge[0]) for edge in pet.in_edges(node.id, EdgeType.CHILD)]
    # add current node to parents, if it is of type FUNC
    if isinstance(node, FunctionNode):
        parents.append(node)
    parent_functions = [cu for cu in parents if isinstance(cu, FunctionNode)]
    for pf in parent_functions:
        for arg in pf.args:
            if var.startswith(arg.name):
                return True
    return False


def is_scalar_val(var) -> bool:
    """Checks if variable is a scalar value

    :param var: variable
    :return: true if scalar
    """
    return not (var.type.endswith("**") or var.type.startswith("ARRAY" or var.type.startswith("[")))


def is_readonly(
    mem_regs: Set[MemoryRegion],
    war: Set[Tuple[NodeID, NodeID, Dependency]],
    waw: Set[Tuple[NodeID, NodeID, Dependency]],
    rev_war: Set[Tuple[NodeID, NodeID, Dependency]],
) -> bool:
    """Checks if variable is readonly

    :param var: variable name
    :param war: war dependencies of the loop
    :param waw: waw dependencies of the loop
    :param rev_war: reversed raw dependencies of the loop
    :return: trie if readonly
    """
    for e in itertools.chain(war, waw, rev_war):
        if e[2].memory_region in mem_regs:
            return False
    return True


def is_global(var: str, tree: Sequence[Node]) -> bool:
    """Checks if variable is global

    :param var: variable name
    :param tree: nodes to search
    :return: true if global
    """

    for node in tree:
        if isinstance(node, CUNode):
            for gv in node.global_vars:
                if gv.name == var:
                    # TODO from tmp global vars
                    return True
    return False


def is_first_written(
    mem_regs: Set[MemoryRegion],
    raw: Set[Tuple[NodeID, NodeID, Dependency]],
    war: Set[Tuple[NodeID, NodeID, Dependency]],
    sub: List[CUNode],
) -> bool:
    """Checks whether a variable is first written inside the current node

    :param var: variable name
    :param raw: raw dependencies of the loop
    :param war: war dependencies of the loop
    :param sub: subtree of the loop
    :return: true if first written
    """
    for e in war:
        if e[2].memory_region in mem_regs and any([n.id == e[1] for n in sub]):
            res = False
            for eraw in raw:
                if (
                    eraw[2].memory_region in mem_regs
                    and any([n.id == e[1] for n in sub])
                    and e[2].source_line == eraw[2].sink_line
                ):
                    res = True
                    break
            if not res:
                return False
    return True


def is_first_written_new(
    var: Variable,
    mem_regs: Set[MemoryRegion],
    raw_deps: Set[Tuple[NodeID, NodeID, Dependency]],
    war_deps: Set[Tuple[NodeID, NodeID, Dependency]],
    reverse_raw_deps: Set[Tuple[NodeID, NodeID, Dependency]],
    reverse_war_deps: Set[Tuple[NodeID, NodeID, Dependency]],
    tree: Sequence[Node],
):
    """Checks whether a variable is first written inside the current node

    :param var:
    :param raw_deps: raw dependencies of the loop
    :param war_deps: war dependencies of the loop
    :param reverse_raw_deps:
    :param reverse_war_deps:
    :param tree: subtree of the loop
    :return: true if first written
    """
    result = False
    # None may occur because __get_variables doesn't check for actual elements
    if var.name is None:
        return False
    is_read = is_read_in(mem_regs, raw_deps, war_deps, reverse_raw_deps, reverse_war_deps, tree)
    if var.name is None:
        print("Empty var.name found. Skipping.")
        return False
    for dep in raw_deps:
        assert dep[2].var_name is not None
        if dep[2].memory_region in mem_regs and any([n.id == dep[1] for n in tree]):
            result = True
            for warDep in war_deps:
                assert warDep[2].var_name is not None
                if (
                    warDep[2].memory_region in mem_regs
                    and any([n.id == dep[1] for n in tree])
                    and dep[2].source_line == warDep[2].sink_line
                ):
                    result = False
                    break
    return result or not is_read


def is_read_in_subtree(
    mem_regs: Set[MemoryRegion], rev_raw: Set[Tuple[NodeID, NodeID, Dependency]], tree: List[Node]
) -> bool:
    """Checks if variable is read in subtree

    :param var: variable name
    :param rev_raw: reversed raw dependencies of the loop
    :param tree: subtree
    :return: true if read in right subtree
    """
    for e in rev_raw:
        if e[2].memory_region in mem_regs and any([n.id == e[0] for n in tree]):
            return True
    return False


def is_read_in_right_subtree(
    mem_regs: Set[MemoryRegion], rev_raw: Set[Tuple[NodeID, NodeID, Dependency]], tree: List[CUNode]
) -> bool:
    """Checks if variable is read in subtree

    :param var: variable name
    :param rev_raw: reversed raw dependencies of the loop
    :param tree: subtree
    :return: true if read in right subtree
    """
    for e in rev_raw:
        if e[2].memory_region in mem_regs:
            if any([n.id == e[1] for n in tree]):
                if not any([k.id == e[0] for k in tree]):
                    return True
    return False


def is_depend_in_out(
    mem_regs: Set[MemoryRegion],
    in_deps: List[Tuple[NodeID, NodeID, Dependency]],
    out_deps: List[Tuple[NodeID, NodeID, Dependency]],
) -> bool:
    """there is an in and out dependency

    :param var: Variable
    :param in_deps: in dependencies
    :param out_deps: out dependencies
    :return: true if dependency is both in and out
    """
    for in_dep in in_deps:
        for out_dep in out_deps:
            if in_dep[2].memory_region in mem_regs and in_dep[2].memory_region == out_dep[2].memory_region:
                return True
    return False


def is_depend_in_var(
    mem_regs: Set[MemoryRegion],
    in_deps: List[Tuple[NodeID, NodeID, Dependency]],
    raw_deps_on: Set[Tuple[NodeID, NodeID, Dependency]],
) -> bool:
    """Checks if variable is written inside a dependent task and read in current task

    :param var: Variable
    :param in_deps: in dependencies
    :param raw_deps_on: raw dependencies
    :return: true if variable is in dependency
    """
    for in_dep in in_deps:
        if in_dep[2].memory_region in mem_regs and in_dep in raw_deps_on:
            return True
    return False


def is_depend_out_var(
    mem_regs: Set[MemoryRegion],
    reverse_raw_deps_on: Set[Tuple[NodeID, NodeID, Dependency]],
    out_deps: List[Tuple[NodeID, NodeID, Dependency]],
) -> bool:
    """Checks if variable is written inside a current task and read in dependent task

    :param var: Variable
    :param reverse_raw_deps_on: raw dependencies
    :param out_deps: in dependencies
    :return: true if variable is out dependency
    """
    for dep in out_deps:
        if dep[2].memory_region in mem_regs and dep in reverse_raw_deps_on:
            return True
    return False


def is_read_in(
    mem_regs: Set[MemoryRegion],
    raw_deps_on: Set[Tuple[NodeID, NodeID, Dependency]],
    war_deps_on: Set[Tuple[NodeID, NodeID, Dependency]],
    reverse_raw_deps_on: Set[Tuple[NodeID, NodeID, Dependency]],
    reverse_war_deps_on: Set[Tuple[NodeID, NodeID, Dependency]],
    tree: Sequence[Node],
) -> bool:
    """Check all reverse RAW dependencies (since we know that var is written in loop, because
    is_first_written returned true)

    :param var: variable
    :param raw_deps_on: raw dependencies
    :param war_deps_on: war dependencies
    :param reverse_raw_deps_on: reverse raw dependencies
    :param reverse_war_deps_on: reverse war dependencies
    :param tree: nodes of the loop
    :return:
    """
    for dep in raw_deps_on:
        # If there is a reverse raw dependency for var and the sink cu is not part
        # of the loop, then var is read in rst
        if dep[2].memory_region in mem_regs:
            return True
    for dep in war_deps_on:
        if dep[2].memory_region in mem_regs and any([n.id == dep[1] for n in tree]):
            return True
    for dep in reverse_raw_deps_on:
        # If there is a reverse raw dependency for var and the sink cu is not part
        # of the loop, then var is read in rst
        if dep[2].memory_region in mem_regs and any([n.id == dep[1] for n in tree]):
            return True
    for dep in reverse_war_deps_on:
        if dep[2].memory_region in mem_regs:
            return True
    return False


def get_child_loops(pet: PEGraphX, node: Node) -> Tuple[List[Node], List[Node]]:
    """Gets all do-all and reduction subloops

    :param pet: CU graph
    :param node: root node
    :return: list of do-all and list of reduction loop nodes
    """
    do_all: List[Node] = []
    reduction: List[Node] = []

    for loop_child in pet.subtree_of_type(node, LoopNode):
        if loop_child.do_all:
            do_all.append(loop_child)
        elif loop_child.reduction:
            reduction.append(loop_child)

    for func_child in pet.direct_children_or_called_nodes_of_type(node, FunctionNode):
        for child in pet.direct_children_or_called_nodes_of_type(func_child, CUNode):
            if child.do_all:
                do_all.append(child)
            elif child.reduction:
                reduction.append(child)

    return do_all, reduction


def get_initialized_memory_regions_in(pet: PEGraphX, cu_nodes: List[CUNode]) -> Dict[Variable, Set[MemoryRegion]]:
    initialized_memory_regions: Dict[Variable, Set[MemoryRegion]] = dict()
    for cu in cu_nodes:
        for s, t, d in pet.out_edges(cu.id, EdgeType.DATA):
            if d.dtype == DepType.INIT and d.memory_region is not None:
                # get variable object from cu
                for var in cu.global_vars + cu.local_vars:
                    if var.name == d.var_name:
                        if var not in initialized_memory_regions:
                            initialized_memory_regions[var] = set()
                        # create entry for initialized variable
                        initialized_memory_regions[var].add(d.memory_region)

    return initialized_memory_regions


def classify_loop_variables(
    pet: PEGraphX, loop: Node
) -> Tuple[List[Variable], List[Variable], List[Variable], List[Variable], List[Variable]]:
    """Classifies variables inside the loop

    :param pet: CU graph
    :param loop: loop node
    :return: first_private, private, last_private, shared, reduction
    """
    first_private = []
    private = []
    last_private = []
    shared = []
    reduction = []
    lst = pet.get_left_right_subtree(loop, False)
    rst = pet.get_left_right_subtree(loop, True)
    sub: List[CUNode] = pet.subtree_of_type(loop, CUNode)

    raw = set()
    war = set()
    waw = set()
    rev_raw = set()

    for sub_node in sub:
        raw.update(__get_dep_of_type(pet, sub_node, DepType.RAW, False))
        war.update(__get_dep_of_type(pet, sub_node, DepType.WAR, False))
        waw.update(__get_dep_of_type(pet, sub_node, DepType.WAW, False))
        rev_raw.update(__get_dep_of_type(pet, sub_node, DepType.RAW, True))

    vars = pet.get_undefined_variables_inside_loop(loop)

    # only consider memory regions which are know at the current code location.
    # ignore memory regions which stem from called functions.
    left_subtree_without_called_nodes = pet.get_left_right_subtree(loop, False, ignore_called_nodes=True)
    prior_known_vars = pet.get_variables(left_subtree_without_called_nodes)
    prior_known_mem_regs = set()
    for pkv in prior_known_vars:
        prior_known_mem_regs.update(prior_known_vars[pkv])
    initialized_in_loop = get_initialized_memory_regions_in(pet, sub)
    initialized_memory_regions: Set[MemoryRegion] = set()
    for var in initialized_in_loop:
        initialized_memory_regions.update(initialized_in_loop[var])

    for var in vars:
        vars[var] = set(
            [
                mem_reg
                for mem_reg in vars[var]
                if mem_reg in prior_known_mem_regs or mem_reg in initialized_memory_regions
            ]
        )

    # vars = list(pet.get_variables(sub))
    for var in vars:
        if is_loop_index2(pet, loop, var.name):
            if is_read_in_subtree(vars[var], rev_raw, rst):
                last_private.append(var)
            else:
                private.append(var)
        elif loop.reduction and pet.is_reduction_var(loop.start_position(), var.name):
            var.operation = pet.get_reduction_sign(loop.start_position(), var.name)
            reduction.append(var)
        elif is_written_in_subtree(vars[var], raw, waw, lst) or is_func_arg(pet, var.name, loop) and is_scalar_val(var):
            if is_readonly(vars[var], war, waw, rev_raw):
                if is_global(var.name, sub):
                    shared.append(var)
                else:
                    first_private.append(var)
            elif is_read_in_subtree(vars[var], rev_raw, rst):
                if is_scalar_val(var):
                    last_private.append(var)
                else:
                    shared.append(var)
            else:
                if not is_scalar_val(var):
                    # array type variable is written
                    shared.append(var)
                else:
                    private.append(var)

        elif is_first_written(vars[var], raw, war, sub):
            if is_read_in_subtree(vars[var], rev_raw, rst):
                if is_scalar_val(var):
                    last_private.append(var)
                else:
                    shared.append(var)
            else:
                if is_scalar_val(var):
                    private.append(var)
                else:
                    shared.append(var)
    # return first_private, private, last_private, shared, reduction
    return (
        sorted(first_private),
        sorted(private),
        sorted(last_private),
        sorted(shared),
        sorted(reduction),
    )


def classify_task_vars(
    pet: PEGraphX,
    task: Node,
    type: str,
    in_deps: List[Tuple[NodeID, NodeID, Dependency]],
    out_deps: List[Tuple[NodeID, NodeID, Dependency]],
    used_in_task_parallelism_detection=False,
) -> Tuple[List[Variable], List[Variable], List[Variable], List[Variable], List[Variable], List[Variable], List[str],]:
    """Classify task variables

    :param pet: CU graph
    :param task: node
    :param type: type of task
    :param in_deps: in dependencies
    :param out_deps: out dependencies
    :param used_in_task_parallelism_detection: set True, if called in a task-parallelism detection context
    """

    first_private: List[Tuple[Variable, Set[MemoryRegion]]] = []
    private: List[Tuple[Variable, Set[MemoryRegion]]] = []
    shared: List[Tuple[Variable, Set[MemoryRegion]]] = []
    depend_in: List[Tuple[Variable, Set[MemoryRegion]]] = []
    depend_out: List[Tuple[Variable, Set[MemoryRegion]]] = []
    depend_in_out: List[Tuple[Variable, Set[MemoryRegion]]] = []
    reduction_var_names: List[str] = []

    left_sub_tree = pet.get_left_right_subtree(task, False)

    right_sub_tree = pet.get_left_right_subtree(task, True)
    subtree = pet.subtree_of_type(task, CUNode)
    t_loop = pet.subtree_of_type(task, LoopNode)

    vars: Dict[Variable, Set[MemoryRegion]] = dict()
    if isinstance(task, FunctionNode):
        tmp = pet.get_variables(subtree)
        vars_strings = []
        for v in task.args:
            vars_strings.append(v.name)
        for v in tmp:
            # None may occur because __get_variables doesn't check for actual elements
            if v.name is None:
                continue

            if "." in v.name:
                name = v.name[0 : v.name.index(".")]  # substring before '.'
            else:
                name = v.name

            if name in vars_strings:
                vars[v] = tmp[v]
    else:
        vars = pet.get_variables(subtree)

    raw_deps_on = set()  # set<Dependence>
    war_deps_on = set()
    waw_deps_on = set()

    reverse_raw_deps_on = set()
    reverse_war_deps_on = set()
    reverse_waw_deps_on = set()

    for sub_node in subtree:
        # insert all entries from child_cu.RAW_deps_on into RAW_deps_on etc.
        raw_deps_on.update(__get_dep_of_type(pet, sub_node, DepType.RAW, False))
        war_deps_on.update(__get_dep_of_type(pet, sub_node, DepType.WAR, False))
        waw_deps_on.update(__get_dep_of_type(pet, sub_node, DepType.WAW, False))

        reverse_raw_deps_on.update(__get_dep_of_type(pet, sub_node, DepType.RAW, True))
        reverse_war_deps_on.update(__get_dep_of_type(pet, sub_node, DepType.WAR, True))
        reverse_waw_deps_on.update(__get_dep_of_type(pet, sub_node, DepType.WAW, True))

    do_all_loops, reduction_loops = get_child_loops(pet, task)
    # reduction_result = ""

    if isinstance(task, LoopNode):
        if task.reduction:
            reduction_loops.append(task)
        else:
            do_all_loops.append(task)

    loop_nodes: List[Node] = [n for n in t_loop if n.reduction]
    if task.reduction:
        loop_nodes.append(task)

    loops_start_lines = [n.start_position() for n in loop_nodes]
    loop_children = [c for n in loop_nodes for c in pet.direct_children_or_called_nodes(n)]

    for var in vars:
        var_is_loop_index = False
        # get RAW dependencies for var
        tmp_deps = [dep for dep in raw_deps_on if dep[2].memory_region in vars[var]]
        for edge in tmp_deps:
            if pet.is_loop_index(edge[2].var_name, loops_start_lines, loop_children):
                var_is_loop_index = True
                break
        if var_is_loop_index:
            private.append((var, vars[var]))
        elif ("GeometricDecomposition" in type or "Pipeline" in type) and is_reduction_any(
            loops_start_lines, var.name, pet.reduction_vars
        ):
            reduction_var_names.append(var.name)
        elif is_depend_in_out(vars[var], in_deps, out_deps):
            depend_in_out.append((var, vars[var]))
        elif is_depend_in_var(vars[var], in_deps, raw_deps_on):
            depend_in.append((var, vars[var]))
        elif is_depend_out_var(vars[var], reverse_raw_deps_on, out_deps):
            depend_out.append((var, vars[var]))
        elif (
            is_written_in_subtree(vars[var], raw_deps_on, waw_deps_on, left_sub_tree)
            or (is_func_arg(pet, var.name, task) and is_scalar_val(var))
        ) and is_readonly(vars[var], war_deps_on, waw_deps_on, reverse_raw_deps_on):
            if is_global(var.name, subtree):
                shared.append((var, vars[var]))
            else:
                first_private.append((var, vars[var]))
        elif is_first_written_new(
            var,
            vars[var],
            raw_deps_on,
            war_deps_on,
            reverse_raw_deps_on,
            reverse_war_deps_on,
            subtree,
        ):
            if is_scalar_val(var) and (
                not used_in_task_parallelism_detection or not __is_written_prior_to_task(pet, var, task)
            ):
                if is_read_in(
                    vars[var],
                    raw_deps_on,
                    war_deps_on,
                    reverse_raw_deps_on,
                    reverse_war_deps_on,
                    right_sub_tree,
                ):
                    shared.append((var, vars[var]))
                else:
                    private.append((var, vars[var]))
            else:
                shared.append((var, vars[var]))

    # use known variables to reconstruct the correct variable names from the classified memory regions
    left_subtree_without_called_nodes = pet.get_left_right_subtree(task, False, ignore_called_nodes=True)
    prior_known_vars = pet.get_variables(left_subtree_without_called_nodes)

    return (
        sorted(__apply_dealiasing(first_private, prior_known_vars)),
        sorted(__apply_dealiasing(private, prior_known_vars)),
        sorted(__apply_dealiasing(shared, prior_known_vars)),
        sorted(__apply_dealiasing(depend_in, prior_known_vars)),
        sorted(__apply_dealiasing(depend_out, prior_known_vars)),
        sorted(__apply_dealiasing(depend_in_out, prior_known_vars)),
        sorted(reduction_var_names),
    )


def __apply_dealiasing(
    input_list: List[Tuple[Variable, Set[MemoryRegion]]],
    previously_known: Dict[Variable, Set[MemoryRegion]],
) -> List[Variable]:
    tmp_memory_regions = set()
    for _, mem_regs in input_list:
        tmp_memory_regions.update(mem_regs)
    cleaned = [pkv for pkv in previously_known if len(previously_known[pkv].intersection(tmp_memory_regions))]
    return cleaned


def __is_written_prior_to_task(pet: PEGraphX, var: Variable, task: Node) -> bool:
    """Check if var has been written in predecessor of task.

    :param pet: CU graph
    :param var: variable
    :param task: node
    """
    # get predecessors of task
    queue: List[Node] = [task]
    visited: List[Node] = []
    predecessors: List[Node] = []
    while queue:
        current = queue.pop()
        if current not in visited:
            visited.append(current)
        if current not in predecessors:
            predecessors.append(current)
        queue += [
            pet.node_at(edge[0])
            for edge in pet.in_edges(current.id, EdgeType.SUCCESSOR)
            if pet.node_at(edge[0]) not in visited
        ]

    # check if raw-dependency on var to any predecessor exists)
    for out_dep in pet.out_edges(task.id, EdgeType.DATA):
        if out_dep[2].dtype != DepType.RAW:
            continue
        # check if out_dep.source in predecessors
        if pet.node_at(out_dep[0]) in predecessors:
            return True
    return False



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import List, Set, Dict, Tuple, Optional

from discopop_explorer.PEGraphX import (
    PEGraphX,
    EdgeType,
    NodeID,
    FunctionNode,
    MemoryRegion,
)


def get_contained_lines(start_line: str, end_line: str) -> List[str]:
    """Returns a list of line numbers inbetween start_line and end_line"""
    file_id = start_line.split(":")[0]
    if file_id != end_line.split(":")[0]:
        raise ValueError("File-ids not equal! ", start_line, end_line)
    line_numbers: List[int] = list(range(int(start_line.split(":")[1]), int(end_line.split(":")[1]) + 1))
    result = [file_id + ":" + str(num) for num in line_numbers]
    return result


def get_function_body_cus_without_called_functions(pet: PEGraphX, function_node: FunctionNode) -> List[NodeID]:
    queue = [t for s, t, d in pet.out_edges(function_node.id, EdgeType.CHILD)]
    visited: Set[NodeID] = set()
    while queue:
        current = queue.pop(0)
        visited.add(current)
        current_node = pet.node_at(current)

        # add children if they do not result from a call
        children = [t for s, t, d in pet.out_edges(current, EdgeType.CHILD)]
        called = [t for s, t, d in pet.out_edges(current, EdgeType.CALLSNODE)]
        queue += [c for c in children if c not in visited and c not in called]  # todo add check for call
    return list(visited)


def prepare_liveness_metadata(
    pet: PEGraphX,
    liveness: Dict[MemoryRegion, List[NodeID]],
    writes: Dict[MemoryRegion, Set[Tuple[NodeID, Optional[int]]]],
    meta_liveness: Dict[MemoryRegion, List[str]],
):
    for mem_reg in liveness:
        for cu_id in liveness[mem_reg]:
            if mem_reg not in meta_liveness:
                meta_liveness[mem_reg] = []
            cu_node = pet.node_at(cu_id)
            write_identifiers: Set[int] = set()
            for cid, ident in writes[mem_reg]:
                if cid == cu_id:
                    if ident is not None:
                        write_identifiers.add(ident)

            for line in get_contained_lines(cu_node.start_position(), cu_node.end_position()):
                if len(write_identifiers) > 0:
                    # is written
                    line = line + ":" + "(" + ",".join([str(ident) for ident in write_identifiers]) + ")"
                else:
                    # is read
                    line = line + ":"
                meta_liveness[mem_reg].append(line)

    # remove duplicates
    for mem_reg in meta_liveness:
        meta_liveness[mem_reg] = list(set(meta_liveness[mem_reg]))
    # if read and write entries for the same line number exist, remove the read access
    for mem_reg in meta_liveness:
        to_be_removed: Set[str] = set()
        for entry in meta_liveness[mem_reg]:
            if ":(" in entry:
                # is written
                continue
            # else, check if write exists in list
            for entry_2 in meta_liveness[mem_reg]:
                split_line_1 = entry.split(":")
                split_line_2 = entry_2.split(":")
                if split_line_1[0] == split_line_2[0] and split_line_1[1] == split_line_2[1]:
                    if split_line_1[2] == split_line_2[2]:
                        continue
                    if split_line_2[2].startswith("("):
                        # split_line_2 is a write access.
                        # remove read entry
                        to_be_removed.add(entry)
        for entry in to_be_removed:
            if entry in meta_liveness[mem_reg]:
                meta_liveness[mem_reg].remove(entry)

    return meta_liveness



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import List, Tuple, cast, Any, Union, Optional

import networkx as nx  # type: ignore

from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.classes.nodes.ContextMerge import ContextMerge
from discopop_library.discopop_optimizer.classes.nodes.ContextRestore import ContextRestore
from discopop_library.discopop_optimizer.classes.nodes.ContextSave import ContextSave
from discopop_library.discopop_optimizer.classes.nodes.ContextSnapshot import ContextSnapshot
from discopop_library.discopop_optimizer.classes.nodes.ContextSnapshotPop import ContextSnapshotPop
from discopop_library.discopop_optimizer.classes.nodes.Loop import Loop
from discopop_library.discopop_optimizer.classes.nodes.Workload import Workload
from discopop_library.discopop_optimizer.utilities.MOGUtilities import (
    get_children,
    get_successors,
)
from discopop_library.discopop_optimizer.utilities.simple_utilities import data_at


# define aliases for type checking purposes
class RegularCosts(Tuple[int, int]):
    pass


class BranchCosts(RegularCosts):
    pass


class WorkloadStack(object):
    stack: List[Union[List[Any], RegularCosts]]  # List of WorkloadStacks or integer tuples (min_wl, max_wl)

    def __init__(self):
        self.stack = []

    def enter_new_branched_section(self):
        """add a new list to the end of self.stack"""
        innermost_stack, _ = self.__get_innermost_workload_stack()
        innermost_stack.append([])

    def exit_branched_section(self):
        """aggregate the contained branches"""
        # do nothing, as the functionality is implemented in aggregate
        pass

    def enter_new_branch(self):
        """same behavior as enter_branched_section"""
        innermost_stack, _ = self.__get_innermost_workload_stack()
        innermost_stack.append([])

    def exit_branch(self):
        """Accumulates costs of the current branch, i.e. the innermost List"""
        # get innermost stack
        innermost_stack, indices = self.__get_innermost_workload_stack()
        # accumulate innermost stack
        min_wl = 0
        max_wl = 0
        for entry in innermost_stack:
            entry_min_wl, entry_max_wl = entry
            min_wl += entry_min_wl
            max_wl += entry_max_wl

        # replace innermost stack with BranchCosts element
        cur_elem = self.stack
        if len(indices) > 0:
            for index in indices[:-1]:
                cur_elem = cur_elem[index]
            cur_elem[indices[-1]] = BranchCosts((min_wl, max_wl))

    def aggregate(self):
        # get innermost stack
        innermost_stack, indices = self.__get_innermost_workload_stack()
        min_wl = 0
        max_wl = 0
        # sum up regular costs and save branch costs in a list for later accumulation
        branch_costs: List[BranchCosts] = []
        for entry in innermost_stack:
            if isinstance(entry, BranchCosts):
                branch_costs.append(entry)
            elif isinstance(entry, RegularCosts):
                min_wl += entry[0]
                max_wl += entry[1]
            else:
                raise ValueError("Unsupported element type: " + str(type(entry)))

        # accumulate branch costs (identify minimum and maximum possible costs)
        min_branch_wl: Optional[int] = None
        max_branch_wl: Optional[int] = None
        for costs in branch_costs:
            min_costs, max_costs = costs
            if min_branch_wl is None:
                min_branch_wl = min_costs
            if max_branch_wl is None:
                max_branch_wl = max_costs
            if min_branch_wl > min_costs:
                min_branch_wl = min_costs
            if max_branch_wl < max_costs:
                max_branch_wl = max_costs

        # accumulate total costs for branched section
        min_wl += 0 if min_branch_wl is None else cast(int, min_branch_wl)
        max_wl += 0 if max_branch_wl is None else cast(int, max_branch_wl)

        # replace innermost stack with accumulated costs
        cur_elem = self.stack
        if len(indices) > 0:
            for index in indices[:-1]:
                cur_elem = cur_elem[index]
            cur_elem[indices[-1]] = RegularCosts((min_wl, max_wl))

    def register_workload(self, min_wl: int, max_wl: int):
        # get the innermost workload stack
        innermost_stack, _ = self.__get_innermost_workload_stack()
        innermost_stack.append(RegularCosts((min_wl, max_wl)))

    def get_min_workload(self) -> int:
        """sum up minimal workloads in the stack"""
        min_wl_sum = 0
        for entry in self.stack:
            if not isinstance(entry, RegularCosts):
                raise ValueError("Unsupported element type: " + str(type(entry)))
            min_wl_sum += entry[0]
        return min_wl_sum

    def get_max_workload(self) -> int:
        """sum up maximal workloads in the stack"""
        max_wl_sum = 0
        for entry in self.stack:
            if not isinstance(entry, RegularCosts):
                raise ValueError("Unsupported element type: " + str(type(entry)))
            max_wl_sum += entry[1]
        return max_wl_sum

    def __get_innermost_workload_stack(self) -> Tuple[List[Any], List[int]]:
        """identifies and returns a reference to the currently active (innermost) workload stack stored in self.stack,
        as well as the list of indices to access this element"""
        indices: List[int] = []
        innermost_stack = self.stack
        while True:
            if len(innermost_stack) == 0:
                break
            if isinstance(innermost_stack[-1], list):
                # save index
                indices.append(len(innermost_stack) - 1)
                # enter the last entry of the currently visited stack and proceed wit the search
                innermost_stack = cast(Any, innermost_stack[-1])

            else:
                # registered workload (tuple of integers) found
                break
        return innermost_stack, indices

    def __get_parent_of_innermost_stack(self) -> Optional[List[Any]]:
        """identifies and returns a reference to the parent List of the innermost workload stack"""
        parents: List[Optional[List[Any]]] = [None]
        innermost_stack = self.stack
        while True:
            if len(innermost_stack) == 0:
                break
            if isinstance(innermost_stack[-1], list):
                # update the parent before entering a new entry
                parents.append(innermost_stack)
                # enter the last entry of the currently visited stack and proceed wit the search
                innermost_stack = innermost_stack[-1]
            else:
                # registered workload (tuple of integers) found
                break
        return parents[-1]


def get_workload_delta_for_cu_node(experiment: Experiment, node_id: int) -> Tuple[int, int, int]:
    """Estimates the expected deviation between different branches in the given node and returns the resulting delta
    as well as the total min / max workloads.
    Unit: same as workload"""
    # get specified node and get delta_wl from specified node
    workload_stack = __parse_node(experiment, node_id, WorkloadStack())
    cu_max_wl = workload_stack.get_max_workload()
    cu_min_wl = workload_stack.get_min_workload()
    delta_workload = workload_stack.get_max_workload() - workload_stack.get_min_workload()

    return delta_workload, cu_min_wl, cu_max_wl


def __parse_node(
    experiment: Experiment,
    node_id: int,
    workload_stack: WorkloadStack,
    ignore_successors: bool = False,
    iteration_factor: int = 1,
) -> WorkloadStack:
    """Calculates the min / max workload and returns them plus the updated WorkloadStack"""
    node_data = data_at(experiment.optimization_graph, node_id)

    min_wl, max_wl = 0, 0
    last_visited_node_id = node_id

    # modify the workload_stack if necessary
    #  ContextSnapshot --> enter recursion
    #  ContextRestore --> start enter new branch
    #  ContextSave --> end of branch
    #  ContextMerge --> end of branched section, calculate min / max (aggregate)
    #  ContextSnapshotPop -> exit recursion
    if type(node_data) == ContextSnapshot:
        workload_stack.enter_new_branched_section()
    elif type(node_data) == ContextRestore:
        workload_stack.enter_new_branch()
    elif type(node_data) == ContextSave:
        workload_stack.exit_branch()
    elif type(node_data) == ContextMerge:
        workload_stack.aggregate()
    elif type(node_data) == ContextSnapshotPop:
        workload_stack.exit_branched_section()

    # calculate the workload of the current node and register it in the stack
    #  add workload of the current node to min / max
    if isinstance(node_data, Workload):
        min_wl = (node_data.sequential_workload or 0) + (node_data.parallelizable_workload or 0)
        max_wl = min_wl
        # apply iteration factor if requested
        min_wl *= iteration_factor
        max_wl *= iteration_factor
        workload_stack.register_workload(min_wl, max_wl)

    # register children
    if type(node_data) == Loop:
        iteration_factor = node_data.iterations
    else:
        iteration_factor = 1
    for child_id in get_children(experiment.optimization_graph, node_id):
        workload_stack = __parse_node(experiment, child_id, workload_stack, iteration_factor=iteration_factor)

    # set next node
    successors = get_successors(experiment.optimization_graph, node_id)

    # continue calculation with successor of node
    # at this point in the OPT Graphs lifecycle, successors can either be empty or have exactly one entry
    if len(successors) == 0:
        # return if none exists
        return workload_stack

    # get min and max workload for successor
    if not ignore_successors:
        workload_stack = __parse_node(experiment, successors[0], workload_stack)

    return workload_stack



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import Optional, List, Any

from discopop_library.CodeGenerator.classes.Enums import PragmaPosition


class Pragma(object):
    pragma_str: str = ""
    file_id: Optional[int] = None
    start_line: Optional[int] = None
    end_line: Optional[int] = None
    pragma_position: PragmaPosition = PragmaPosition.BEFORE_START
    parent_cu_id: str = ""
    children: List[Any] = []

    def __init__(self):
        self.children = []  # create individual list for each Pragma object to prevent bugs due to mutability



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import tkinter as tk
from typing import Dict, cast, List, Tuple, Optional, Any

import jsonpickle  # type: ignore
import networkx as nx  # type: ignore
import sympy  # type: ignore
from spb import plot3d, MB  # type: ignore
from sympy import Integer, Expr, Symbol, lambdify, plot, Float, init_printing, simplify, diff  # type: ignore

from discopop_library.discopop_optimizer.CostModels.CostModel import CostModel
from discopop_library.discopop_optimizer.CostModels.DataTransfer.DataTransferCosts import (
    add_data_transfer_costs,
)
from discopop_library.discopop_optimizer.CostModels.utilities import (
    get_performance_models_for_functions,
)
from discopop_library.discopop_optimizer.DataTransfers.DataTransfers import calculate_data_transfers
from discopop_library.discopop_optimizer.PETParser.PETParser import PETParser
from discopop_library.discopop_optimizer.Variables.Experiment import Experiment
from discopop_library.discopop_optimizer.Variables.ExperimentUtils import (
    show_function_models,
    export_to_json,
    perform_headless_execution,
)
from discopop_library.discopop_optimizer.classes.context.ContextObject import ContextObject
from discopop_library.discopop_optimizer.classes.enums.Distributions import FreeSymbolDistribution
from discopop_library.discopop_optimizer.classes.nodes.FunctionRoot import FunctionRoot
from discopop_library.discopop_optimizer.gui.queries.ValueTableQuery import (
    query_user_for_symbol_values,
)
from discopop_library.discopop_optimizer.suggestions.importers.main import import_suggestions
from discopop_library.discopop_optimizer.utilities.optimization.LocalOptimization.TopDown import (
    get_locally_optimized_models,
)


class OptimizationGraph(object):
    next_free_node_id: int

    def __init__(
        self,
        experiment: Experiment,
        arguments: Dict[str, Any],
        parent_frame: Optional[tk.Frame],
        destroy_window_after_execution: bool,
    ):
        # construct optimization graph from PET Graph
        # save graph in experiment
        experiment.optimization_graph, self.next_free_node_id = PETParser(experiment).parse()

        # get performance models for sequential execution
        sequential_function_performance_models = get_performance_models_for_functions(
            experiment, experiment.optimization_graph
        )
        sequential_function_performance_models_with_transfers = calculate_data_transfers(
            experiment.optimization_graph, sequential_function_performance_models, experiment
        )
        sequential_complete_performance_models = add_data_transfer_costs(
            experiment.optimization_graph,
            sequential_function_performance_models_with_transfers,
            experiment,
        )

        # import parallelization suggestions
        experiment.optimization_graph = import_suggestions(experiment)

        # perform an exhaustive search if requested
        if arguments["--exhaustive-search"]:
            # calculate performance models without data transfers
            function_performance_models = get_performance_models_for_functions(
                experiment, experiment.optimization_graph
            )

            # calculate and append necessary data transfers to the models
            function_performance_models_with_transfers = calculate_data_transfers(
                experiment.optimization_graph, function_performance_models, experiment
            )

            # calculate and append costs of data transfers to the performance models
            exhaustive_performance_models = add_data_transfer_costs(
                experiment.optimization_graph,
                function_performance_models_with_transfers,
                experiment,
            )
        else:
            exhaustive_performance_models = dict()

        # collect free symbols
        #        free_symbols: Set[Symbol] = set()
        free_symbol_ranges: Dict[Symbol, Tuple[float, float]] = dict()
        free_symbol_distributions: Dict[Symbol, FreeSymbolDistribution] = dict()

        #        for function in complete_performance_models:
        #            for pair in complete_performance_models[function]:
        #                model, context = pair
        #                free_symbols.update(cast(List[Symbol], model.parallelizable_costs.free_symbols))
        #                free_symbols.update(cast(List[Symbol], model.sequential_costs.free_symbols))
        #                suggested_values = suggested_values | model.symbol_value_suggestions
        sorted_free_symbols = sorted(list(experiment.free_symbols), key=lambda x: x.name)

        # query user for values for free symbols
        query_results = query_user_for_symbol_values(sorted_free_symbols, experiment.suggested_values)
        for symbol, value, start_value, end_value, symbol_distribution in query_results:
            if value is not None:
                experiment.substitutions[symbol] = Float(value)
            else:
                free_symbol_ranges[symbol] = (cast(float, start_value), cast(float, end_value))
                free_symbol_distributions[symbol] = cast(FreeSymbolDistribution, symbol_distribution)

        # by default, select the sequential version of each function for substitution
        for function in sequential_complete_performance_models:
            experiment.selected_paths_per_function[function] = sequential_complete_performance_models[function][0]

        # add function symbols to list of substitutions
        # collect substitutions
        for function in experiment.selected_paths_per_function:
            # register substitution
            experiment.substitutions[cast(Symbol, function.sequential_costs)] = experiment.selected_paths_per_function[
                function
            ][0].sequential_costs
            experiment.substitutions[
                cast(Symbol, function.parallelizable_costs)
            ] = experiment.selected_paths_per_function[function][0].parallelizable_costs

        # TODO END OF DUMMY

        # set free symbol ranges and distributions for comparisons
        #        for idx, function in enumerate(complete_performance_models):
        #            for pair in complete_performance_models[function]:
        #                model, context = pair
        #                model.free_symbol_ranges = free_symbol_ranges
        #                model.free_symbol_distributions = free_symbol_distributions

        # save substitutions, sorted_free_symbols, free_symbol_ranges and free_symbol_distributions in experiment
        experiment.sorted_free_symbols = sorted_free_symbols
        experiment.free_symbol_ranges = free_symbol_ranges
        experiment.free_symbol_distributions = free_symbol_distributions

        # create locally optimized model
        locally_optimized_models = get_locally_optimized_models(
            experiment,
            experiment.optimization_graph,
            experiment.substitutions,
            experiment,
            free_symbol_ranges,
            free_symbol_distributions,
        )

        # set free symbol ranges and distributions for comparisons
        for idx, function in enumerate(locally_optimized_models):
            for pair in locally_optimized_models[function]:
                model, context = pair
                model.free_symbol_ranges = free_symbol_ranges
                model.free_symbol_distributions = free_symbol_distributions

        # set free symbol ranges and distributions for comparisons
        for idx, function in enumerate(exhaustive_performance_models):
            for pair in exhaustive_performance_models[function]:
                model, context = pair
                model.free_symbol_ranges = free_symbol_ranges
                model.free_symbol_distributions = free_symbol_distributions

        # find the minimum of the exhaustive list of models
        exhaustive_minima: Dict[FunctionRoot, List[Tuple[CostModel, ContextObject]]] = dict()
        if len(exhaustive_performance_models) > 0:
            print("Sorting exhaustive models...")
            for function in exhaustive_performance_models:
                sorted_exhaustive_performance_models = sorted(
                    exhaustive_performance_models[function], key=lambda x: x[0]
                )
                exhaustive_minima[function] = [sorted_exhaustive_performance_models[0]]
            print("\tDone.")

        for function in sequential_complete_performance_models:
            # show table of options
            options: List[Tuple[CostModel, ContextObject, str]] = []
            # add exhaustive minima
            if function in exhaustive_minima:
                for model_tmp in exhaustive_minima[function]:
                    options.append((model_tmp[0], model_tmp[1], "Exhaustive Min"))

            options.append(
                (
                    sequential_complete_performance_models[function][0][0],
                    sequential_complete_performance_models[function][0][1],
                    "Sequential",
                )
            )
            if function in locally_optimized_models:
                options.append(
                    (
                        locally_optimized_models[function][0][0],
                        locally_optimized_models[function][0][1],
                        "Locally Optimized",
                    )
                )
            # save options to experiment
            experiment.function_models[function] = options

        # show function models
        if not arguments["--headless-mode"]:
            if parent_frame is None:
                raise ValueError("No frame provided!")
            show_function_models(experiment, parent_frame, destroy_window_after_execution)
        else:
            # perform actions for headless mode
            perform_headless_execution(experiment)

    def get_next_free_node_id(self):
        buffer = self.next_free_node_id
        self.next_free_node_id += 1
        return buffer



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import copy
from typing import List, Tuple, Optional, cast, Dict

from discopop_explorer.PEGraphX import Node, NodeType, EdgeType, PEGraphX
from discopop_explorer.pattern_detectors.PatternInfo import PatternInfo
from discopop_explorer.pattern_detectors.task_parallelism.classes import (
    TaskParallelismInfo,
    ParallelRegionInfo,
    OmittableCuInfo,
    TPIType,
)
from discopop_explorer.pattern_detectors.task_parallelism.tp_utils import (
    get_parent_of_type,
    task_contained_in_reduction_loop,
)


def suggest_parallel_regions(pet: PEGraphX, suggestions: List[TaskParallelismInfo]) -> List[ParallelRegionInfo]:
    """create suggestions for parallel regions based on suggested tasks.
    Parallel regions are suggested aroung each outer-most function call
    possibly leading to the creation of tasks.
    To obtain these, the child-graph is traversed in reverse,
    starting from each suggested task.
    :param pet: PET graph
    :param suggestions: List[TaskParallelismInfo]
    :return: List[ParallelRegionInfo]"""
    # get task suggestions from suggestions
    task_suggestions = [s for s in suggestions if s.type is TPIType.TASK]
    # start search for each suggested task
    parents: List[Tuple[Node, Optional[Node]]] = []
    for ts in task_suggestions:
        parents += get_parent_of_type(pet, ts._node, NodeType.FUNC, EdgeType.CHILD, False)
    # remove duplicates
    parents = list(set(parents))
    # get outer-most parents of suggested tasks
    outer_parents: List[Tuple[Node, Optional[Node]]] = []
    # iterate over entries in parents.
    while len(parents) > 0:
        p, last_node = parents.pop(0)
        p_parents = get_parent_of_type(pet, p, NodeType.FUNC, EdgeType.CHILD, False)
        if not p_parents:  # p_parents is empty
            # p is outer
            # get last cu before p
            outer_parents.append((p, last_node))
        else:
            # append p´s parents to queue, filter out entries if already
            # present in outer_parents
            first_elements = [x[0] for x in outer_parents]
            parents += [x for x in p_parents if x[0] not in first_elements]

    # create region suggestions based on detected outer parents
    region_suggestions = []
    for parent, last_node in outer_parents:
        if last_node is None:
            continue
        last_node = last_node
        region_suggestions.append(
            ParallelRegionInfo(parent, TPIType.PARALLELREGION, last_node.start_position(), last_node.end_position())
        )
    return region_suggestions


def set_task_contained_lines(suggestions: List[TaskParallelismInfo]) -> List[TaskParallelismInfo]:
    """set region_end_line property of TaskParallelismInfo objects
    in suggestions and return the modified list.
    Regions are determined by checking if a CU contains multiple Tasks or
    Barriers and splitting up the contained source code lines accordingly.
    :param suggestions: List[TaskParallelismInfo]
    :return: List[TaskParallelismInfo]"""
    # group suggestions by parent CU
    output = []
    cu_to_suggestions_map: Dict[str, List[TaskParallelismInfo]] = dict()
    for s in suggestions:
        # fill cu_to_suggestions_map
        if s.node_id in cu_to_suggestions_map:
            cu_to_suggestions_map[s.node_id].append(s)
        else:
            cu_to_suggestions_map[s.node_id] = [s]
    # order suggestions for each CU by first affected line
    for cu in cu_to_suggestions_map:
        sorted_suggestions = cu_to_suggestions_map[cu]
        sorted_suggestions.sort(key=lambda x: x.region_start_line)
        cu_to_suggestions_map[cu] = sorted_suggestions
    # iterate over suggestions. set region_end_line to end of cu or
    # beginning of next suggestion
    for cu in cu_to_suggestions_map:
        for idx, s in enumerate(cu_to_suggestions_map[cu]):
            # check if next element exists
            if idx + 1 < len(cu_to_suggestions_map[cu]):
                # if so, set end to line prior to start of next suggestion
                end = int(cu_to_suggestions_map[cu][idx + 1].region_start_line)
                end = end - 1
                s.region_end_line = str(end)
            else:
                # if not, set end to end of cu
                s.region_end_line = s.end_line[s.end_line.index(":") + 1 :]
            # overwrite entry in cu_to_suggestions_map for s
            cu_to_suggestions_map[cu][idx] = s
    # append suggestions to output
    for cu in cu_to_suggestions_map:
        for s in cu_to_suggestions_map[cu]:
            output.append(s)
    return output


def detect_taskloop_reduction(pet: PEGraphX, suggestions: List[TaskParallelismInfo]) -> List[TaskParallelismInfo]:
    """detect suggested tasks which can and should be replaced by
    taskloop reduction.
    return the modified list of suggestions.
    Idea:   1. check if suggested task inside loop body
            2. check if outer loop is reduction loop
                3. if so, build reduction clause and modify suggested task
    :param pet: PET graph
    :param suggestions: List[TaskParallelismInfo]
    :return: List[TaskParallelismInfo]
    """
    output = []
    # iterate over suggestions
    for s in suggestions:
        # ignore others than tasks
        if s.type is not TPIType.TASK:
            continue
        # check if s contained in reduction loop body
        red_vars_entry, red_loop = task_contained_in_reduction_loop(pet, s)
        if red_vars_entry is None or red_loop is None:
            # s not contained in reduction loop body
            output.append(s)
        else:
            red_vars_entry = red_vars_entry
            red_loop = red_loop
            # s contained in reduction loop body
            # modify task s
            reduction_clause = "reduction("
            reduction_clause += red_vars_entry["operation"] + ":"
            reduction_clause += red_vars_entry["name"].replace(".addr", "")
            reduction_clause += ")"
            s.pragma = ["taskloop", reduction_clause]
            s.type = TPIType.TASKLOOP
            # update pragma line to parent reduction loop
            s.pragma_line = red_loop.start_position()
            # update pragma region
            s.region_start_line = red_loop.start_position()
            s.region_end_line = red_loop.end_position()
            # append modified task to output
            output.append(s)
    return output


def combine_omittable_cus(pet: PEGraphX, suggestions: List[PatternInfo]) -> List[PatternInfo]:
    """execute combination of tasks suggestions with omittable cus.
    Adds modified version of the respective Parent suggestions to the list.
    Returns the modified list of suggestions.
    Omittable CU suggetions are removed from the list.
    Removes duplicates in in/out/in_out dependency lists.
    :param pet: PET graph
    :param suggestions: List [PatternInfo]
    :return: List[PatternInfo]
    """
    omittable_suggestions: List[OmittableCuInfo] = []
    task_suggestions: List[TaskParallelismInfo] = []
    result: List[PatternInfo] = []
    for single_suggestion in suggestions:
        if type(single_suggestion) == OmittableCuInfo:
            omittable_suggestions.append(single_suggestion)
        else:
            if type(single_suggestion) == TaskParallelismInfo:
                try:
                    if single_suggestion.type is TPIType.TASK:
                        task_suggestions.append(single_suggestion)
                    else:
                        result.append(single_suggestion)
                except AttributeError:
                    result.append(single_suggestion)
            else:
                result.append(single_suggestion)

    # remove omittable suggestion if cu is no direct child in the
    # successor graph of a node containing a task suggestion
    useful_omittable_suggestions = []
    for oms in omittable_suggestions:
        in_succ_edges = [(s, t, e) for s, t, e in pet.in_edges(oms._node.id) if e.etype == EdgeType.SUCCESSOR]
        parent_task_nodes = [pet.node_at(e[0]) for e in in_succ_edges if pet.node_at(e[0]).tp_contains_task is True]
        if len(parent_task_nodes) != 0:
            useful_omittable_suggestions.append(oms)
        else:
            # un-mark node as omittable
            oms._node.tp_omittable = False
    omittable_suggestions = useful_omittable_suggestions

    # create copies of original Task suggestion versions
    for omit_s in omittable_suggestions:
        for ts in task_suggestions:
            if omit_s.combine_with_node == ts._node:
                result.append(copy.copy(ts))

    # prepare dict to find target suggestions for combination
    task_suggestions_dict: Dict[Node, List[TaskParallelismInfo]] = dict()
    for ts in task_suggestions:
        if ts._node in task_suggestions_dict:
            task_suggestions_dict[ts._node].append(ts)
        else:
            task_suggestions_dict[ts._node] = [ts]

    for omit_s in omittable_suggestions:
        # process in_out dependencies of omit_s
        # -> lazy, let following statements take care
        for omit_in_out_var in omit_s.in_out_dep:
            omit_s.in_dep.append(omit_in_out_var)
            omit_s.out_dep.append(omit_in_out_var)

        # find target task_suggestion for omit_s, based on in / out dep matches
        omit_target_task_indices = []
        if omit_s.combine_with_node in task_suggestions_dict:
            if len(task_suggestions_dict[omit_s.combine_with_node]) != 1:
                # search for matching in/out dependency pair
                for idx, ts in enumerate(task_suggestions_dict[omit_s.combine_with_node]):
                    intersect = [v for v in omit_s.in_dep if v in ts.out_dep]
                    if len(intersect) == len(omit_s.in_dep):
                        # all in_deps covered
                        omit_target_task_indices.append(idx)
            else:
                omit_target_task_indices = [0]

            for omit_target_task_idx in omit_target_task_indices:
                # note: dependencies of task nodes can contain multiples
                # process out dependencies of omit_s
                for omit_out_var in omit_s.out_dep:
                    if omit_out_var is None:
                        continue
                    task_suggestions_dict[omit_s.combine_with_node][omit_target_task_idx].out_dep.append(omit_out_var)
                    # omit_s.combine_with_node.out_dep.append(omit_out_var)
                # process in dependencies of omit_s
                for omit_in_var in omit_s.in_dep:
                    # note: only dependencies to target node allowed
                    if omit_in_var in task_suggestions_dict[omit_s.combine_with_node][omit_target_task_idx].out_dep:
                        task_suggestions_dict[omit_s.combine_with_node][omit_target_task_idx].out_dep.remove(
                            omit_in_var
                        )
                    # omit_s.combine_with_node.out_dep.remove(omit_in_var)

                # increase size of pragma region if needed
                if ":" not in cast(
                    str,
                    task_suggestions_dict[omit_s.combine_with_node][omit_target_task_idx].region_end_line,
                ):
                    if int(omit_s.end_line[omit_s.end_line.index(":") + 1 :]) > int(
                        cast(
                            str,
                            task_suggestions_dict[omit_s.combine_with_node][omit_target_task_idx].region_end_line,
                        )
                    ):
                        task_suggestions_dict[omit_s.combine_with_node][
                            omit_target_task_idx
                        ].region_end_line = omit_s.end_line
                else:
                    cut_region_end_line = cast(
                        str,
                        task_suggestions_dict[omit_s.combine_with_node][omit_target_task_idx].region_end_line,
                    )
                    cut_region_end_line = cut_region_end_line[cut_region_end_line.index(":") + 1 :]
                    if int(omit_s.end_line[omit_s.end_line.index(":") + 1 :]) > int(cut_region_end_line):
                        task_suggestions_dict[omit_s.combine_with_node][
                            omit_target_task_idx
                        ].region_end_line = omit_s.end_line

    # remove duplicates from dependency lists and append to result
    for key in task_suggestions_dict:
        for ts in task_suggestions_dict[key]:
            # remove duplicates
            ts.in_dep = list(set(ts.in_dep))
            ts.out_dep = list(set(ts.out_dep))
            # reset in_out_dep, might have changed due to combination
            if len(ts.in_dep) < len(ts.out_dep):  # just for performance
                ts.in_out_dep = [var for var in ts.in_dep if var in ts.out_dep]
            else:
                ts.in_out_dep = [var for var in ts.out_dep if var in ts.in_dep]
            ts.in_out_dep = list(set(ts.in_out_dep))
            result.append(ts)

    return result



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import sys
from typing import Dict, List, Set, Tuple, cast

import networkx as nx  # type: ignore

from discopop_explorer.PEGraphX import (
    PEGraphX,
    NodeType,
    EdgeType,
    CUNode,
    NodeID,
    MemoryRegion,
    Dependency,
    FunctionNode,
)
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Aliases import (
    VarName,
)
from discopop_explorer.pattern_detectors.combined_gpu_patterns.classes.Enums import UpdateType
from discopop_explorer.pattern_detectors.combined_gpu_patterns.utilities import (
    get_function_body_cus_without_called_functions,
)


def populate_live_data(comb_gpu_reg, pet: PEGraphX, ignore_update_instructions=False) -> Dict[VarName, List[NodeID]]:
    """calculate List of cu-id's in the combined region for each variable in which the respective data is live.
    The gathered information is used for the optimization / creation of data mapping instructions afterwards.
    """
    liveness: Dict[VarName, List[NodeID]] = dict()

    # populate liveness sets based on gpu loops
    for region in comb_gpu_reg.contained_regions:
        for gpu_loop in region.contained_loops:
            live_in_loop = (
                [v.name for v in gpu_loop.map_type_to]
                + [v.name for v in gpu_loop.map_type_tofrom]
                + [v.name for v in gpu_loop.map_type_from]
                + [v.name for v in gpu_loop.map_type_alloc]
                + [v.name for v in gpu_loop.reduction_vars_ids]
            )
            # set liveness within loop
            subtree = pet.subtree_of_type(pet.node_at(gpu_loop.node_id), CUNode)
            for var in live_in_loop:
                if var not in liveness:
                    liveness[var] = []
                for cu in subtree:
                    liveness[var].append(cu.id)

    if not ignore_update_instructions:
        # populate liveness sets based on update instructions
        for source_id, sink_id, update_type, var, meta_line_num in comb_gpu_reg.update_instructions:
            if var not in liveness:
                liveness[var] = []
            if update_type == UpdateType.TO_DEVICE:
                liveness[var].append(sink_id)
            elif update_type == UpdateType.FROM_DEVICE:
                liveness[var].append(source_id)
            else:
                raise ValueError("Unsupported Update type: ", update_type)

    return liveness


def add_memory_regions_to_device_liveness(
    device_liveness: Dict[VarName, List[NodeID]],
    cu_and_variable_to_memory_regions: Dict[NodeID, Dict[VarName, Set[MemoryRegion]]],
) -> Dict[VarName, List[Tuple[NodeID, Set[MemoryRegion]]]]:
    extended_device_liveness: Dict[VarName, List[Tuple[NodeID, Set[MemoryRegion]]]] = dict()

    for var_name in device_liveness:
        if var_name not in extended_device_liveness:
            extended_device_liveness[var_name] = []
        for cu_id in device_liveness[var_name]:
            memory_regions: Set[MemoryRegion] = set()
            if cu_id in cu_and_variable_to_memory_regions:
                if var_name in cu_and_variable_to_memory_regions[cu_id]:
                    memory_regions.update(cu_and_variable_to_memory_regions[cu_id][var_name])
            extended_device_liveness[var_name].append((cu_id, memory_regions))

    return extended_device_liveness


def propagate_memory_regions(
    device_liveness_plus_memory_regions: Dict[VarName, List[Tuple[NodeID, Set[MemoryRegion]]]]
) -> Dict[VarName, List[Tuple[NodeID, Set[MemoryRegion]]]]:
    """Propagate memory regions for variables"""
    result_dict: Dict[VarName, List[Tuple[NodeID, Set[MemoryRegion]]]] = dict()
    for var_name in device_liveness_plus_memory_regions:
        cu_ids: List[NodeID] = []
        memory_regions: Set[MemoryRegion] = set()
        for cu_id, mem_regs in device_liveness_plus_memory_regions[var_name]:
            cu_ids.append(cu_id)
            memory_regions.update(mem_regs)

        if var_name not in result_dict:
            result_dict[var_name] = []
        for cu_id in cu_ids:
            result_dict[var_name].append((cu_id, memory_regions))

    return result_dict


def convert_liveness(
    device_liveness_plus_memory_regions: Dict[VarName, List[Tuple[NodeID, Set[MemoryRegion]]]]
) -> Dict[MemoryRegion, List[NodeID]]:
    result_dict: Dict[MemoryRegion, List[NodeID]] = dict()

    for var_name in device_liveness_plus_memory_regions:
        for cu_id, mem_regions in device_liveness_plus_memory_regions[var_name]:
            for mem_reg in mem_regions:
                if mem_reg not in result_dict:
                    result_dict[mem_reg] = []
                result_dict[mem_reg].append(cu_id)
    return result_dict


def extend_data_lifespan(
    pet: PEGraphX, live_data: Dict[MemoryRegion, List[NodeID]]
) -> Dict[MemoryRegion, List[NodeID]]:
    """Extends the lifespan of the data on the device to allow as little data movement as possible."""
    print("Extending data lifespan...", file=sys.stderr)
    # create graph copy and remove all but successor edges
    copied_graph = pet.g.copy()

    # remove all but successor edges
    to_be_removed = set()
    for edge in copied_graph.edges:
        edge_data = cast(Dependency, copied_graph.edges[edge]["data"])
        if edge_data.etype != EdgeType.SUCCESSOR:
            to_be_removed.add(edge)
    for edge in to_be_removed:
        copied_graph.remove_edge(edge[0], edge[1])

    for idx, mem_reg in enumerate(live_data):
        print("EXTENDING mem_reg ", idx, "/", len(live_data), file=sys.stderr)
        modification_found = True
        cycles = 0

        finished_functions: Set[FunctionNode] = set()

        while modification_found:
            print("\t", "len: ", len(live_data[mem_reg]), file=sys.stderr)
            cycles += 1
            modification_found = False
            new_entries: List[Tuple[MemoryRegion, NodeID]] = []

            path_node_ids: Set[NodeID] = set()
            path_nodes: Set[CUNode] = set()

            # split cu_ids by their parent function (used to reduce the complexity of the following step
            cu_ids_by_parent_functions: Dict[FunctionNode, List[NodeID]] = dict()
            for cu_id in live_data[mem_reg]:
                parent_function = pet.get_parent_function(pet.node_at(cu_id))
                if parent_function not in cu_ids_by_parent_functions:
                    cu_ids_by_parent_functions[parent_function] = []
                cu_ids_by_parent_functions[parent_function].append(cu_id)

            print("\tparent functions: ", len(cu_ids_by_parent_functions), file=sys.stderr)
            print(
                "\t",
                [(key.name, len(cu_ids_by_parent_functions[key])) for key in cu_ids_by_parent_functions],
                file=sys.stderr,
            )

            for parent_function in cu_ids_by_parent_functions:
                # removed due to incorrect results!
                #                if parent_function in finished_functions:
                #                    print("\t\tskipped finished function: ", parent_function.name, file=sys.stderr)
                #                    continue
                function_new_entries: List[Tuple[MemoryRegion, NodeID]] = []
                new_path_node_found = False
                for cu_id in cu_ids_by_parent_functions[parent_function]:
                    # check if data is live in any successor
                    # If so, set mem_reg to live in each of the encountered CUs.

                    for potential_successor_cu_id in cu_ids_by_parent_functions[parent_function]:
                        if cu_id == potential_successor_cu_id:
                            continue

                        if cu_id in parent_function.reachability_pairs:
                            reachable = potential_successor_cu_id in parent_function.reachability_pairs[cu_id]
                        else:
                            reachable = False

                        if reachable:
                            # get nodes of path from source to target
                            for path in nx.all_simple_paths(
                                copied_graph, source=cu_id, target=potential_successor_cu_id
                            ):
                                len_pre = len(path_node_ids)
                                path_node_ids.update(path)
                                if len_pre < len(path_node_ids):
                                    new_path_node_found = True

                    # set mem_reg to live in every child of CU
                    for _, child_id, _ in pet.out_edges(cu_id, EdgeType.CHILD):
                        if child_id not in live_data[mem_reg]:
                            function_new_entries.append((mem_reg, child_id))

                    # set mem_reg to live in every called function of CU
                    for _, called_node, _ in pet.out_edges(cu_id, EdgeType.CALLSNODE):
                        if called_node not in live_data[mem_reg]:
                            function_new_entries.append((mem_reg, called_node))

                new_entries += function_new_entries

                function_finished = (
                    not new_path_node_found
                )  # len(function_new_entries) == 0 and not new_path_node_found
                if function_finished:
                    finished_functions.add(parent_function)

            path_nodes.update([cast(CUNode, pet.node_at(nid)) for nid in path_node_ids])

            # if path_node is located within a loop, add the other loop cus to the path as well
            to_be_added: List[CUNode] = []
            for path_node in path_nodes:
                parent_node = [pet.node_at(s) for s, t, d in pet.in_edges(path_node.id, EdgeType.CHILD)][0]
                if parent_node.type == NodeType.LOOP:
                    for _, loop_cu_id, _ in pet.out_edges(parent_node.id, EdgeType.CHILD):
                        loop_cu = cast(CUNode, pet.node_at(loop_cu_id))
                        if loop_cu not in path_nodes and loop_cu not in to_be_added:
                            to_be_added.append(loop_cu)
            for loop_cu in to_be_added:
                path_nodes.add(loop_cu)

            # mark mem_reg live in all path_nodes and their children
            for path_node in path_nodes:
                # todo replace with subtree calculation after merging with refactoring changes
                # calculate subtree without including called functions
                subtree_without_called_functions = [
                    cu
                    for cu in pet.direct_children(path_node)
                    if cu not in [pet.node_at(t) for s, t, d in pet.out_edges(path_node.id, EdgeType.CALLSNODE)]
                ]
                # add path_node itself to the subtree
                subtree_without_called_functions.append(path_node)
                # todo end of section to be replaced
                #  subtree = pet.subtree_of_type(path_node, NodeType.CU)  # subtree contains path_node
                for subtree_node in subtree_without_called_functions:
                    if subtree_node.id not in [cu_id for cu_id in live_data[mem_reg]]:
                        new_entries.append((mem_reg, subtree_node.id))

            new_entries = list(set(new_entries))
            if len(new_entries) > 0:
                modification_found = True
            for mem_reg, new_cu_id in new_entries:
                live_data[mem_reg].append(new_cu_id)

        print("\tCycles: ", cycles, file=sys.stderr)

    # remove duplicates
    for mem_reg in live_data:
        live_data[mem_reg] = list(set(live_data[mem_reg]))

    print("\tDone.", file=sys.stderr)
    return live_data


def calculate_host_liveness(
    comb_gpu_reg,
    pet: PEGraphX,
) -> Dict[VarName, List[Tuple[NodeID, Set[MemoryRegion]]]]:
    """
    Variable is live on host, if a dependency between the host cu or any of its children and
    any device cu for a given variable exists

    """

    host_liveness_lists: Dict[VarName, List[Tuple[NodeID, Set[MemoryRegion]]]] = dict()

    all_function_cu_ids: Set[NodeID] = set()
    for region in comb_gpu_reg.contained_regions:
        parent_function = pet.get_parent_function(pet.node_at(region.node_id))
        all_function_cu_ids.update(get_function_body_cus_without_called_functions(pet, parent_function))

    all_function_host_cu_ids = [cu_id for cu_id in all_function_cu_ids if cu_id not in comb_gpu_reg.device_cu_ids]

    for cu_id in all_function_host_cu_ids:
        shared_variables: Set[VarName] = set()
        shared_memory_regions: Set[MemoryRegion] = set()
        # get all data which is accessed by the cu_id and it's children and any device cu
        subtree = pet.subtree_of_type(pet.node_at(cu_id), CUNode)
        for subtree_node_id in [n.id for n in subtree]:
            out_data_edges = pet.out_edges(subtree_node_id, EdgeType.DATA)
            in_data_edges = pet.in_edges(subtree_node_id, EdgeType.DATA)
            for _, target, dep in out_data_edges:
                if target in comb_gpu_reg.device_cu_ids:
                    if dep.var_name is not None:
                        shared_variables.add(VarName(dep.var_name))
                    if dep.memory_region is not None:
                        shared_memory_regions.add(MemoryRegion(dep.memory_region))

            for source, _, dep in in_data_edges:
                if source in comb_gpu_reg.device_cu_ids:
                    if dep.var_name is not None:
                        shared_variables.add(VarName(dep.var_name))
                    if dep.memory_region is not None:
                        shared_memory_regions.add(dep.memory_region)
        for var_name in shared_variables:
            if var_name not in host_liveness_lists:
                host_liveness_lists[var_name] = []
            host_liveness_lists[var_name].append((cu_id, shared_memory_regions))

    # convert sets to lists
    host_liveness: Dict[VarName, List[Tuple[NodeID, Set[MemoryRegion]]]] = dict()
    for key in host_liveness_lists:
        host_liveness[key] = host_liveness_lists[key]
    return host_liveness



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.


from multiprocessing import Pool
from typing import List, cast

from alive_progress import alive_bar  # type: ignore

from .PatternInfo import PatternInfo
from ..PEGraphX import (
    CUNode,
    LoopNode,
    PEGraphX,
    NodeType,
    Node,
    LineID,
    DepType,
    EdgeType,
)
from ..utils import is_reduction_var, classify_loop_variables
from ..variable import Variable


class ReductionInfo(PatternInfo):
    """Class, that contains reduction detection result"""

    def __init__(self, pet: PEGraphX, node: Node):
        """
        :param pet: PET graph
        :param node: node, where reduction was detected
        """
        PatternInfo.__init__(self, node)
        self.pragma = "#pragma omp parallel for"

        fp, p, lp, s, r = classify_loop_variables(pet, node)
        self.first_private = fp
        self.private = p
        self.last_private = lp
        self.shared = s
        self.reduction = r

    def __str__(self):
        return (
            f"Reduction at: {self.node_id}\n"
            f"Start line: {self.start_line}\n"
            f"End line: {self.end_line}\n"
            f"pragma: {self.pragma}\n"
            f"private: {[v.name for v in self.private]}\n"
            f"shared: {[v.name for v in self.shared]}\n"
            f"first private: {[v.name for v in self.first_private]}\n"
            f'reduction: {[v.operation + ":" + v.name for v in self.reduction]}\n'
            f"last private: {[v.name for v in self.last_private]}"
        )


global_pet = None


def run_detection(pet: PEGraphX) -> List[ReductionInfo]:
    """Search for reduction pattern

    :param pet: PET graph
    :return: List of detected pattern info
    """
    import tqdm  # type: ignore

    global global_pet
    global_pet = pet
    result: List[ReductionInfo] = []
    nodes = pet.all_nodes(LoopNode)

    param_list = [(node) for node in nodes]
    with Pool(initializer=__initialize_worker, initargs=(pet,)) as pool:
        tmp_result = list(tqdm.tqdm(pool.imap_unordered(__check_node, param_list), total=len(param_list)))
    for local_result in tmp_result:
        result += local_result
    print("GLOBAL RES: ", result)

    for pattern in result:
        pattern.get_workload(pet)
        pattern.get_per_iteration_workload(pet)

    return result


def __initialize_worker(pet):
    global global_pet
    global_pet = pet


def __check_node(param_tuple):
    global global_pet
    local_result = []
    node = param_tuple
    if __detect_reduction(global_pet, node):
        node.reduction = True
        if node.loop_iterations >= 0 and not node.contains_array_reduction:
            local_result.append(ReductionInfo(global_pet, node))

    return local_result


def __detect_reduction(pet: PEGraphX, root: LoopNode) -> bool:
    """Detects reduction pattern in loop

    :param pet: PET graph
    :param root: the loop node
    :return: true if is reduction loop
    """
    all_vars = []
    for node in pet.subtree_of_type(root, CUNode):
        all_vars.extend(node.local_vars)
        all_vars.extend(node.global_vars)

    # get required metadata
    loop_start_lines: List[LineID] = []
    root_children = pet.subtree_of_type(root, (CUNode, LoopNode))
    root_children_cus: List[CUNode] = [cast(CUNode, cu) for cu in root_children if cu.type == NodeType.CU]
    root_children_loops: List[LoopNode] = [cast(LoopNode, cu) for cu in root_children if cu.type == NodeType.LOOP]
    for v in root_children_loops:
        loop_start_lines.append(v.start_position())
    reduction_vars = [
        v
        for v in all_vars
        if is_reduction_var(root.start_position(), v.name, pet.reduction_vars)
        # and "**" not in v.type --> replaced by check for array reduction
    ]
    reduction_var_names = [v.name for v in reduction_vars]
    fp, p, lp, s, r = classify_loop_variables(pet, root)

    if __check_loop_dependencies(
        pet,
        root,
        root_children_cus,
        root_children_loops,
        loop_start_lines,
        reduction_var_names,
        fp,
        p,
        lp,
    ):
        return False

    # mark loop as containing array reductions, if variable types are accordingly
    if reduction_vars and len([v for v in reduction_vars if "**" in v.type]) > 0:
        root.contains_array_reduction = True

    # if the loop contains any reduction variable, create a reduction suggestion
    return bool(reduction_vars)


def __check_loop_dependencies(
    pet: PEGraphX,
    root_loop: LoopNode,
    root_children_cus: List[CUNode],
    root_children_loops: List[LoopNode],
    loop_start_lines: List[LineID],
    reduction_var_names: List[str],
    first_privates: List[Variable],
    privates: List[Variable],
    last_privates: List[Variable],
) -> bool:
    """Returns True, if dependencies between the respective subgraphs chave been found.
    Returns False otherwise, which results in the potential suggestion of a Reduction pattern."""
    # get recursive children of source and target
    loop_children_ids = [node.id for node in root_children_cus]

    # get dependency edges between children nodes
    deps = set()
    for n in loop_children_ids:
        deps.update([(s, t, d) for s, t, d in pet.in_edges(n, EdgeType.DATA) if s in loop_children_ids])
        deps.update([(s, t, d) for s, t, d in pet.out_edges(n, EdgeType.DATA) if t in loop_children_ids])

    for source, target, dep in deps:
        # check if targeted variable is readonly inside loop
        if pet.is_readonly_inside_loop_body(
            dep,
            root_loop,
            root_children_cus,
            root_children_loops,
            loops_start_lines=loop_start_lines,
        ):
            # variable is readonly -> no problem
            continue

        # check if targeted variable is loop index
        if pet.is_loop_index(dep.var_name, loop_start_lines, root_children_cus):
            continue

        # targeted variable is not read-only
        if dep.dtype == DepType.INIT:
            continue
        elif dep.dtype == DepType.RAW:
            # check RAW dependencies
            # Reductions are only valid, if the value of the reduction variable is not stored in a shared variable.
            # This property is violated if a RAW dependency for the reduction variable between different CUs exist
            # since CUs follow the Read-Compute-Write pattern.
            if dep.var_name in reduction_var_names:
                if source != target:
                    # if raw_deps for reduction variables between different CU's exist,
                    # the above described property is violated
                    # --> not a valid reduction
                    return True
            else:
                # RAW does not target a reduction variable.
                # RAW problematic, if it is not an intra-iteration RAW.
                if not dep.intra_iteration:
                    return True
        elif dep.dtype == DepType.WAR:
            # check WAR dependencies
            # WAR problematic, if it is not an intra-iteration WAR and the variable is not private or firstprivate
            if not dep.intra_iteration:
                if dep.var_name not in [v.name for v in first_privates + privates + last_privates]:
                    return True
        elif dep.dtype == DepType.WAW:
            # check WAW dependencies
            # handled by variable classification
            pass
        else:
            raise ValueError("Unsupported dependency type: ", dep.dtype)

    # no problem found. Potentially suggest reduction
    return False



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import os.path
from argparse import ArgumentParser
from pathlib import Path

from discopop_library.PatchGenerator.PatchGeneratorArguments import PatchGeneratorArguments
from discopop_library.PatchGenerator.patch_generator import run
from discopop_library.ConfigProvider.config_provider import run as run_config_provider
from discopop_library.ConfigProvider.ConfigProviderArguments import ConfigProviderArguments


def parse_args() -> PatchGeneratorArguments:
    """Parse the arguments passed to the discopop_explorer"""
    parser = ArgumentParser(description="DiscoPoP Patch Generator")
    # all flags that are not considered stable should be added to the experimental_parser
    experimental_parser = parser.add_argument_group(
        "EXPERIMENTAL",
        "Arguments marked as experimental features. These flags may or may not be removed or changed in the future.",
    )

    # fmt: off
    parser.add_argument("-v", "--verbose", action="store_true",
        help="Enable verbose output.")
    parser.add_argument(
        "-a", "--add-from-json", type=str, default="None",
        help="Add additional patches specified in the given patterns.json file."
    )
    # EXPERIMENTAL FLAGS:
    # fmt: on

    arguments = parser.parse_args()

    # determine DP build path
    arguments.dp_build_path = run_config_provider(
        ConfigProviderArguments(return_dp_build_dir=True, return_dp_source_dir=False, return_llvm_bin_dir=False)
    )

    # determine LLVM_BIN_DIR
    llvm_bin_dir = run_config_provider(
        ConfigProviderArguments(return_dp_build_dir=False, return_dp_source_dir=False, return_llvm_bin_dir=True)
    )
    # determine CC
    if os.path.exists(os.path.join(llvm_bin_dir, "clang")):
        arguments.cc = os.path.join(llvm_bin_dir, "clang")
    elif os.path.exists(os.path.join(llvm_bin_dir, "clang-11")):
        arguments.cc = os.path.join(llvm_bin_dir, "clang-11")
    else:
        raise ValueError("Could not determine CC from LLVM_BIN_DIR: ", llvm_bin_dir)

    # determine CXX
    if os.path.exists(os.path.join(llvm_bin_dir, "clang++")):
        arguments.cxx = os.path.join(llvm_bin_dir, "clang++")
    elif os.path.exists(os.path.join(llvm_bin_dir, "clang++-11")):
        arguments.cxx = os.path.join(llvm_bin_dir, "clang++-11")
    else:
        raise ValueError("Could not determine CXX from LLVM_BIN_DIR: ", llvm_bin_dir)

    return PatchGeneratorArguments(
        verbose=arguments.verbose,
        discopop_build_path=arguments.dp_build_path,
        CC=arguments.cc,
        CXX=arguments.cxx,
        add_from_json=arguments.add_from_json,
    )


def main():
    arguments = parse_args()
    run(arguments)


if __name__ == "__main__":
    main()



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import tkinter as tk
from tkinter import *
from typing import List, Tuple, Optional, Dict, cast, Union

from sympy import Symbol, Expr

from discopop_library.discopop_optimizer.classes.enums.Distributions import FreeSymbolDistribution
from discopop_library.discopop_optimizer.gui.widgets.ScrollableFrame import ScrollableFrameWidget


def query_user_for_symbol_values(
    symbols: List[Symbol],
    suggested_values: Dict[Symbol, Expr],
) -> List[Tuple[Symbol, Optional[float], Optional[float], Optional[float], Optional[FreeSymbolDistribution]]]:
    """Opens a GUI-Table to query values for each given Symbol from the user.
    The queried values are: Specific value, Range start, Range end.
    In every case, either a specific value, or a range must be given.
    In case the optimizer is started in headless mode, the suggested values are used.
    Return: [(symbol, symbol_value, range_start, range_end, distribution)]"""
    query_result: List[
        Tuple[
            Symbol,
            Optional[float],
            Optional[float],
            Optional[float],
            Optional[FreeSymbolDistribution],
        ]
    ] = []
    column_headers = ["Symbol Name", "Symbol Value", "Range Start", "Range End", "Range Relevance"]

    root = Tk()
    parent_frame = Frame(root)

    # configure weights
    parent_frame.rowconfigure(0, weight=1)
    parent_frame.columnconfigure(0, weight=1)
    # create scrollable frame
    scrollable_frame_widget = ScrollableFrameWidget(parent_frame)
    scrollable_frame = scrollable_frame_widget.get_scrollable_frame()

    rows = []
    # set column headers
    header_cols = []
    for col_idx, header in enumerate(column_headers):
        e = Entry(scrollable_frame, relief=RIDGE)
        e.grid(row=0, column=col_idx, sticky=NSEW)
        e.insert(END, header)
        e.configure(state=DISABLED, disabledforeground="black")
        header_cols.append(e)
    rows.append(header_cols)

    # create choice vars for range relevance
    range_relevance_vars: Dict[Symbol, tk.StringVar] = dict()
    for free_symbol in symbols:
        range_relevance_vars[free_symbol] = tk.StringVar()

    # create query Table
    for row_idx, free_symbol in enumerate(symbols):
        row_idx = row_idx + 1  # to account for column headers
        cols = []
        for col_idx, column_name in enumerate(column_headers):
            e = Entry(scrollable_frame, relief=RIDGE)
            e.grid(row=row_idx, column=col_idx, sticky=NSEW)
            if col_idx == 0:
                # symbol name
                e.insert(END, str(free_symbol))
                e.configure(state=DISABLED, disabledforeground="black")
            elif col_idx == 4:
                # range relevance
                choices = ["-->", "<--", "=="]
                range_relevance_vars[free_symbol].set(choices[0])
                option_menu = tk.OptionMenu(e, range_relevance_vars[free_symbol], *choices)
                option_menu.grid(sticky=NSEW)
            else:
                # queried value
                if col_idx == 1 and free_symbol in suggested_values:
                    # insert suggested value, if one exists
                    e.insert(END, str(suggested_values[free_symbol]))
            cols.append(e)
        rows.append(cols)
    root.mainloop()

    def validate() -> bool:
        ret_val = True
        # validate entries
        for row_idx, row in enumerate(rows):
            if row_idx == 0:
                continue
            row_valid = False
            # check if either a specific value or ranges are set
            if len(row[1].get()) != 0:
                # specific value
                row_valid = True
            elif len(row[2].get()) != 0 and len(row[3].get()) != 0:
                # range specified
                row_valid = True
            else:
                # row invalid
                ret_val = False
            # mark row
            if row_valid:
                for col in row[1:]:
                    col.configure(background="grey")
            else:
                for col in row[1:]:
                    col.configure(background="red")
        return ret_val

    def onPress():
        if not validate():
            return
        # fetch entries
        for row_idx, row in enumerate(rows):
            if row_idx == 0:
                continue
            row_element = []
            for col_idx, col in enumerate(row):
                if col_idx == 0:
                    # append symbol to row_element
                    row_element.append(symbols[row_idx - 1])  # -1 to account for column headers
                elif col_idx == 4:
                    # ignore range relevance, as it is added afterwards
                    pass
                else:
                    field_value = col.get()
                    if len(field_value) == 0:
                        row_element.append(None)
                    else:
                        row_element.append(float(field_value))
            # get enum object from range relevance choice if no specific value has been set in the row
            if row_element[1] is None:
                string_value = range_relevance_vars[row_element[0]].get()
                if string_value == "-->":
                    range_relevance = FreeSymbolDistribution.RIGHT_HEAVY
                elif string_value == "<--":
                    range_relevance = FreeSymbolDistribution.LEFT_HEAVY
                else:
                    range_relevance = FreeSymbolDistribution.UNIFORM
            else:
                range_relevance = None
            row_element.append(range_relevance)

            # create result tuple
            query_result.append(
                cast(
                    Tuple[
                        Symbol,
                        Optional[float],
                        Optional[float],
                        Optional[float],
                        Optional[FreeSymbolDistribution],
                    ],
                    tuple(row_element),
                )
            )

        # close elements on optimizer_frame
        for c in parent_frame.winfo_children():
            c.destroy()

        parent_frame.quit()

    scrollable_frame_widget.finalize(len(symbols), row=0, col=0)
    Button(parent_frame, text="Save", command=lambda: onPress()).grid()
    Button(parent_frame, text="Validate", command=lambda: validate()).grid()
    parent_frame.mainloop()
    return query_result



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from enum import Enum
from typing import List, Set, cast, Tuple

from discopop_explorer.PEGraphX import (
    PEGraphX,
    CUNode,
    DepType,
    NodeID,
    Node,
    DummyNode,
    FunctionNode,
    LoopNode,
    Dependency,
)
from discopop_explorer.utils import is_func_arg, is_global, __get_dep_of_type as get_dep_of_type
from discopop_explorer.variable import Variable


def map_node(pet: PEGraphX, nodeID: NodeID) -> Node:
    return pet.node_at(nodeID)


class map_type_t(Enum):
    MT_TO = 1
    MT_FROM = 2
    MT_TOFROM = 3
    MT_ALLOC = 4
    MT_NONE = 5


def set_find(x: str, s: Set[str]) -> str:
    """implements the cpp function set::find

    :param x: Node to find in set
    :param s: Set to search for given node
    :return: returns either node if it is found, last elem otherwise
    """
    if x in s:
        return x  # found node
    elif not s:
        return ""  # given set empty
    else:
        return s.pop()  # returns last elem of set


def set_end(s: Set[str]) -> str:
    """implements the cpp function set::end

    :param s: set of Node IDs
    :return: last elem of given set
    """
    if not s:
        return ""  # given set empty
    else:
        return s.pop()  # returns last elem of set without popping


def getCalledFunctions(
    pet: PEGraphX, node: Node, calledFunctions: Set[NodeID], dummyFunctions: Set[NodeID]
) -> Set[NodeID]:
    """This function traverses all children nodes of 'node' and adds every
        encountered function (non-dummy) to the 'calledFunctions' set.
        Dummy functions (e.g. functions from linked libraries) are added to the
        'dummyFunctions' set.
        (not recursive anymore, but called function subtree_of_type is)

    :param pet:
    :param node:
    :param calledFunctions:
    :param dummyFunctions:
    :return:
    """
    sub_func = pet.subtree_of_type(node, FunctionNode)
    for f in sub_func:
        calledFunctions.add(f.id)
    # unnecessary in this i think
    sub_dummy = pet.subtree_of_type(node, DummyNode)
    for d in sub_dummy:
        dummyFunctions.add(d.id)

    return calledFunctions


def getDeps(
    cuIDs: List[CUNode],
    pet: PEGraphX,
    RAWDepsOn: Set[Tuple[NodeID, NodeID, Dependency]],
    WARDepsOn: Set[Tuple[NodeID, NodeID, Dependency]],
    WAWDepsOn: Set[Tuple[NodeID, NodeID, Dependency]],
    reverseRAWDepsOn: Set[Tuple[NodeID, NodeID, Dependency]],
    reverseWARDepsOn: Set[Tuple[NodeID, NodeID, Dependency]],
    reverseWAWDepsOn: Set[Tuple[NodeID, NodeID, Dependency]],
) -> None:
    """gather all dependencies of the nodes specified in 'cuIDs'
        ---doesnt work here because addresses---

    :param pet:
    :param cuIDs:
    :param RAWDepsOn:
    :param WARDepsOn:
    :param WAWDepsOn:
    :param reverseRAWDepsOn:
    :param reverseWARDepsOn:
    :param reverseWAWDepsOn:
    :return:
    """
    for sub_node in cuIDs:
        # insert all entries from child_cu.RAW_deps_on into RAW_deps_on etc.
        RAWDepsOn.update(get_dep_of_type(pet, sub_node, DepType.RAW, False))
        WARDepsOn.update(get_dep_of_type(pet, sub_node, DepType.WAR, False))
        WAWDepsOn.update(get_dep_of_type(pet, sub_node, DepType.WAW, False))
        reverseRAWDepsOn.update(get_dep_of_type(pet, sub_node, DepType.RAW, True))
        reverseWARDepsOn.update(get_dep_of_type(pet, sub_node, DepType.WAR, True))
        reverseWAWDepsOn.update(get_dep_of_type(pet, sub_node, DepType.WAW, True))


def assignMapType(
    loop: CUNode,
    loopCUs: List[NodeID],
    var: Variable,
    isScalar: bool,
    pet: PEGraphX,
    RAWDepsOn: Set[Tuple[NodeID, NodeID, Dependency]],
    reverseRAWDepsOn: Set[Tuple[NodeID, NodeID, Dependency]],
) -> map_type_t:
    """assigns a map-type to the variable 'var'

    WARDepsOn: Set[Dependency], WAWDepsOn: Set[Dependency], reverseWARDepsOn: Set[Dependency],
    reverseWAWDepsOn: Set[Dependency] all unused
    :param loop:
    :param loopCUs:
    :param var:
    :param isScalar:
    :param pet:
    :param RAWDepsOn:
    :param reverseRAWDepsOn:
    :return:
    """
    # The dependencies are scanned to detect where data
    # transfers to or from the device are required.
    copy_to: bool = False
    copy_from: bool = False

    # is_global takes list of nodes, not ids
    loop_nodes: List[LoopNode] = []
    for i in loopCUs:
        loop_nodes.append(cast(LoopNode, map_node(pet, i)))  # for later

    # a variable need to be copied to target memory if there exists a RAW
    # dependency for this variable with the sink inside the loop and the source
    # outside of it
    # [Tuple[str, str, Dependency]]:
    for dep in RAWDepsOn:
        if var.name != dep[2].var_name:
            continue

        source_in_loop = False
        if dep[2].var_name in loopCUs:
            source_in_loop = True

        if not source_in_loop:
            copy_to = True
            break

    # a variable needs to be copied to host memory after leaving the loop, if
    # there exists a reverseRAW dependency for this variable with the sink
    # outside the loop and the source inside it
    for dep in reverseRAWDepsOn:
        if var.name != dep[2].var_name:
            continue

        sink_in_loop = False
        if dep[2].var_name in loopCUs:
            sink_in_loop = True

        if not sink_in_loop:
            copy_from = True
            break

    # based on which data transfers are required, decide on which map-type to
    # assign to this variable
    if copy_to and not copy_from:
        return map_type_t.MT_TO
    elif not copy_to and copy_from:
        return map_type_t.MT_FROM
    elif copy_to and copy_from:
        return map_type_t.MT_TOFROM

    # no RAW dependencies exist for function arguments and unmodified global variables
    # - however, since the variable is accessed in the loop assume it is read
    if is_func_arg(pet, var.name, loop) or is_global(var.name, loop_nodes):
        return map_type_t.MT_TO

    # if none of the above conditions are fulfilled, no data transfer is required
    return map_type_t.MT_NONE if isScalar else map_type_t.MT_ALLOC



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
import random
import sys
from typing import List, Dict, Tuple, Optional

import numpy as np
import sympy
from matplotlib import pyplot as plt  # type: ignore
from sympy import Function, Symbol, init_printing, Expr, N, nsimplify, Integer  # type: ignore

from discopop_library.discopop_optimizer.classes.enums.Distributions import FreeSymbolDistribution


class CostModel(object):
    path_decisions: List[int]
    identifier: str
    parallelizable_costs: Expr
    sequential_costs: Expr
    raw_parallelizable_costs: Optional[Expr]
    raw_sequential_costs: Optional[Expr]
    free_symbol_ranges: Dict[Symbol, Tuple[float, float]]
    free_symbol_distributions: Dict[Symbol, FreeSymbolDistribution]
    symbol_value_suggestions: Dict[Symbol, Expr]

    def toJSON(self):
        return "AsDF"

    def __init__(
        self,
        parallelizable_costs: Expr,
        sequential_costs: Expr,
        identifier: str = "None",
        path_decisions=None,
        symbol_value_suggestions: Optional[Dict[Symbol, Expr]] = None,
    ):
        if sequential_costs == sympy.nan:
            raise ValueError("NAN: ", sequential_costs)
        if path_decisions is None:
            self.path_decisions = []
        else:
            # used for the construction of combined models
            self.path_decisions = path_decisions
        if symbol_value_suggestions is None:
            self.symbol_value_suggestions = dict()
        else:
            self.symbol_value_suggestions = symbol_value_suggestions
        self.identifier = identifier
        self.parallelizable_costs = parallelizable_costs
        self.sequential_costs = sequential_costs

    def __str__(self):
        return str(self.parallelizable_costs) + str(self.sequential_costs)

    def print(self, file=sys.stdout):
        init_printing()
        print("\tPARALLEL:")
        print("\t", self.parallelizable_costs, file=file)
        print("\tSERIAL")
        print("\t", self.sequential_costs, file=file)

    def parallelizable_plus_combine(self, other):
        """Combines both models in the following fashion:
        f(x,y) =
        ==> x.parallelizable_costs + y.parallelizable_costs
        ==> x.sequential_costs + y.sequential_costs"""
        if other is None:
            return self
        parallelizable_costs = self.parallelizable_costs + other.parallelizable_costs
        sequential_costs = self.sequential_costs + other.sequential_costs
        path_decisions = self.path_decisions + other.path_decisions
        # merge dictionaries
        value_suggestions = {**self.symbol_value_suggestions, **other.symbol_value_suggestions}
        return CostModel(
            parallelizable_costs,
            sequential_costs,
            path_decisions=path_decisions,
            symbol_value_suggestions=value_suggestions,
        )

    def parallelizable_divide_combine(self, other):
        """Combines both models in the following fashion:
        f(x,y) =
        ==> x.parallelizable_costs / y.parallelizable_costs
        ==> x.sequential_costs / y.sequential_costs"""
        if other is None:
            return self
        parallelizable_costs = self.parallelizable_costs / other.parallelizable_costs
        sequential_costs = self.sequential_costs / other.sequential_costs
        path_decisions = self.path_decisions + other.path_decisions
        value_suggestions = self.symbol_value_suggestions | other.symbol_value_suggestions
        return CostModel(
            parallelizable_costs,
            sequential_costs,
            path_decisions=path_decisions,
            symbol_value_suggestions=value_suggestions,
        )

    def parallelizable_multiply_combine(self, other):
        """Combines both models in the following fashion:
        f(x,y) =
        ==> x.parallelizable_costs * y.parallelizable_costs
        ==> x.sequential_costs * y.sequential_costs"""
        if other is None:
            return self
        parallelizable_costs = self.parallelizable_costs * other.parallelizable_costs
        sequential_costs = self.sequential_costs * other.sequential_costs
        path_decisions = self.path_decisions + other.path_decisions
        # merge dictionaries
        value_suggestions = {**self.symbol_value_suggestions, **other.symbol_value_suggestions}
        return CostModel(
            parallelizable_costs,
            sequential_costs,
            path_decisions=path_decisions,
            symbol_value_suggestions=value_suggestions,
        )

    def register_child(self, other, root_node, experiment, all_function_nodes, current_device):
        """Registers a child node for the given model.
        Does not modify the stored model in self or other."""
        return root_node.register_child(other, experiment, all_function_nodes, current_device)

    def register_successor(self, other, root_node):
        """Registers a successor node for the given model.
        Does not modify the stored model in self or other."""
        return root_node.register_successor(other)

    def __lt__(self, other):
        """Compare both models.
        The comparison is based on random sampling and may not be correct in all cases!
        """
        decision_tendency = 0  # positive -> self was evaluated to be smaller more often than the other way around
        decided = False
        counter = 0
        # Sampling parameters
        min_count = 50
        max_count = 300
        decision_threshold = 0.85
        # Weibull distribution parameters
        alpha, beta = 0.8, 1.3

        # draw and evaluate random samples until either the max_count has been reached, or a decision has been made
        # sampling points may be drawn from a uniform or weibull distribution in a left-skewed or right-skewed
        # configuration. The Parameters of the weibull distribution can be defined above.
        while not (decided or counter > max_count) or counter < min_count:
            counter += 1
            # determine random sampling point
            sampling_point = dict()
            for symbol in self.free_symbol_ranges:
                range_min, range_max = self.free_symbol_ranges[symbol]
                if self.free_symbol_distributions[symbol] == FreeSymbolDistribution.UNIFORM:
                    # draw from uniform distribution
                    sampling_point[symbol] = random.uniform(range_min, range_max)
                else:
                    # use_weibull_distribution
                    # get normalized random value from distribution
                    normalized_pick = 42.0
                    while normalized_pick < 0 or normalized_pick > 1:
                        normalized_pick = random.weibullvariate(alpha, beta)
                    if self.free_symbol_distributions[symbol] == FreeSymbolDistribution.LEFT_HEAVY:
                        # calculate sampling point using the range starting from minimum
                        sampling_point[symbol] = range_min + (range_max - range_min) * normalized_pick
                    else:
                        # simulate a right heavy distribution
                        # calculate sampling point using the range starting from maximum
                        sampling_point[symbol] = range_max - (range_max - range_min) * normalized_pick

            # evaluate both functions at the sampling point
            substituted_model_1_1 = self.parallelizable_costs.xreplace(sampling_point)
            numerical_result_1_1 = substituted_model_1_1.evalf()

            substituted_model_1_2 = self.sequential_costs.xreplace(sampling_point)
            numerical_result_1_2 = substituted_model_1_2.evalf()

            substituted_model_2_1 = other.parallelizable_costs.xreplace(sampling_point)
            numerical_result_2_1 = substituted_model_2_1.evalf()

            substituted_model_2_2 = other.sequential_costs.xreplace(sampling_point)
            numerical_result_2_2 = substituted_model_2_2.evalf()

            # use re() to get real values in case extrap has introduced sqrt's
            total_1 = sympy.re(numerical_result_1_1 + numerical_result_1_2) + sympy.im(
                numerical_result_1_1 + numerical_result_1_2
            )
            total_2 = sympy.re(numerical_result_2_1 + numerical_result_2_2) + sympy.im(
                numerical_result_2_1 + numerical_result_2_2
            )

            # replace Expr(0) with 0
            total_1 = total_1.subs({Expr(Integer(0)): Integer(0)})
            total_2 = total_2.subs({Expr(Integer(0)): Integer(0)})

            # determine relation between the numerical results
            try:
                if total_1 < total_2:
                    decision_tendency += 1
                else:
                    decision_tendency -= 1
            except TypeError as te:
                print("Total 1: ", total_1)
                print("Total 2: ", total_2)
                raise te

            if counter > min_count:
                # check if a decision in either direction can be made
                if decision_threshold < (abs(decision_tendency) / counter):
                    decided = True

        # check if a decision has been made
        if decided:
            if decision_tendency > 0:
                return True
            else:
                return False
        else:
            return False

    def __plot_weibull_distributions(self, alpha: float, beta: float):
        """For Debug reasons. Plots the left and right side heavy weibull distributions using the given parameters."""
        x = np.arange(1, 100.0) / 100.0  # normalized to [0,1]

        def weibull(x, n, a):
            return (a / n) * (x / n) ** (a - 1) * np.exp(-((x / n) ** a))

        plt.plot(x, weibull(x, alpha, beta), label="Alpha: " + str(alpha) + " Beta: " + str(beta))

        # show random picks
        ax = plt.subplot(1, 1, 1)  # type: ignore
        k = 100
        for i in range(0, k):
            # get normalized values
            y_rnd = 42.0
            while y_rnd < 0 or y_rnd > 1:
                y_rnd = random.weibullvariate(alpha, beta)
            ax.plot(y_rnd, 1, "or")
        plt.legend()
        plt.show()



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from enum import Enum
from typing import List, Optional

from discopop_explorer.PEGraphX import Node, MWType, PEGraphX
from discopop_explorer.pattern_detectors.PatternInfo import PatternInfo


# We decided to omit the information that computes the workload and the relevant codes. For large programs (e.g., ffmpeg), the generated Data.xml file becomes very large. However, we keep the code here because we would like to integrate a hotspot detection algorithm (TODO: Bertin) with the parallelism discovery. Then, we need to retrieve the information to decide which code sections (loops or functions) are worth parallelizing.
# from discopop_explorer.utils import total_instructions_count, calculate_workload


class Task(object):
    """This class represents task in task parallelism pattern"""

    nodes: List[Node]
    child_tasks: List["Task"]
    start_line: str
    end_line: str

    def __init__(self, pet: PEGraphX, node: Node):
        self.node_id = node.id
        self.nodes = [node]
        self.start_line = node.start_position()
        if ":" in self.start_line:
            self.region_start_line = self.start_line[self.start_line.index(":") + 1 :]
        else:
            self.region_start_line = self.start_line
        self.region_end_line = None
        self.end_line = node.end_position()
        self.mw_type = node.mw_type

        # We decided to omit the information that computes the workload and the relevant codes. For large programs (e.g., ffmpeg),
        # the generated Data.xml file becomes very large. However, we keep the code here because we would like to integrate a
        # hotspot detection algorithm (TODO: Bertin) with the parallelism discovery.
        # Then, we need to retrieve the information to decide which code sections (loops or functions) are worth parallelizing.
        #        self.instruction_count = total_instructions_count(pet, node)
        #        self.workload = calculate_workload(pet, node)
        self.instruction_count = 0
        self.workload = 0
        self.child_tasks = []

    def aggregate(self, other: "Task"):
        """Aggregates given task into current task

        :param other: task to aggregate
        """
        self.nodes.extend(other.nodes)
        self.end_line = other.end_line
        self.workload += other.workload
        self.instruction_count += other.instruction_count
        self.mw_type = MWType.BARRIER_WORKER if other.mw_type == MWType.BARRIER_WORKER else MWType.WORKER


class TPIType(Enum):
    DUMMY = "DUMMY"
    TASK = "TASK"
    TASKWAIT = "TASKWAIT"
    TASKLOOP = "TASKLOOP"
    PARALLELREGION = "PARALLELREGION"


class TaskParallelismInfo(PatternInfo):
    """Class, that contains task parallelism detection result"""

    def __init__(self, node: Node, type: TPIType, pragma, pragma_line, first_private, private, shared):
        """
        :param node: node, where task parallelism was detected
        :param type: type of the suggestion (task, taskwait, taskloop)
        :param pragma: pragma to be used (task / taskwait / taskloop)
        :param pragma_line: line prior to which the pragma shall be inserted
        :param first_private: list of varNames
        :param private: list of varNames
        :param shared: list of varNames
        """
        PatternInfo.__init__(self, node)
        self.type = type
        self.pragma = pragma
        self.pragma_line = pragma_line
        if ":" in self.pragma_line:
            self.region_start_line = self.pragma_line[self.pragma_line.index(":") + 1 :]
        else:
            self.region_start_line = self.pragma_line
        self.region_end_line: Optional[str] = None
        self.first_private = first_private
        self.private = private
        self.shared = shared
        self.in_dep: List[str] = []
        self.out_dep: List[str] = []
        self.in_out_dep: List[str] = []
        self.critical_sections: List[str] = []
        self.atomic_sections: List[str] = []
        self.task_group: List[int] = []

    def __str__(self):
        return (
            f"Task parallelism at CU: {self.node_id}\n"
            f"CU Start line: {self.start_line}\n"
            f"CU End line: {self.end_line}\n"
            f"pragma at line: {self.pragma_line}\n"
            f"pragma region start line: {self.region_start_line}\n"
            f"pragma region end line: {self.region_end_line}\n"
            f'pragma: "#pragma omp {" ".join(self.pragma)}"\n'
            f'first_private: {" ".join(self.first_private)}\n'
            f'private: {" ".join(self.private)}\n'
            f'shared: {" ".join(self.shared)}\n'
            f'in_dep: {" ".join(self.in_dep)}\n'
            f'out_dep: {" ".join(self.out_dep)}\n'
            f'in_out_dep: {" ".join(self.in_out_dep)}\n'
            f'critical_sections: {" ".join(self.critical_sections)}\n'
            f'atomic_sections: {" ".join(self.atomic_sections)}\n'
            f'task_group: {" ".join([str(e) for e in self.task_group])}\n'
        )


class ParallelRegionInfo(PatternInfo):
    """Class, that contains parallel region info."""

    def __init__(self, node: Node, type: TPIType, region_start_line, region_end_line):
        PatternInfo.__init__(self, node)
        self.region_start_line = region_start_line
        self.region_end_line = region_end_line
        self.pragma = "#pragma omp parallel\n\t#pragma omp single"
        self.type = type

    def __str__(self):
        return (
            f"Task Parallel Region at CU: {self.node_id}\n"
            f"CU Start line: {self.start_line}\n"
            f"CU End line: {self.end_line}\n"
            f"pragma: \n\t{self.pragma}\n"
            f"Parallel Region Start line: {self.region_start_line}\n"
            f"Parallel Region End line {self.region_end_line}\n"
        )


class OmittableCuInfo(PatternInfo):
    """Class, that contains information on omittable CUs (such that can be
    combined with a suggested task).
    Objects of this type are only intermediate and will not show up in the
    final suggestions.
    """

    def __init__(self, node: Node, combine_with_node: Node):
        PatternInfo.__init__(self, node)
        self.combine_with_node = combine_with_node
        # only for printing
        self.cwn_id = combine_with_node.id
        self.in_dep: List[Optional[str]] = []
        self.out_dep: List[Optional[str]] = []
        self.in_out_dep: List[Optional[str]] = []

    def __str__(self):
        return (
            f"Omittable CU: {self.node_id}\n"
            f"CU Start line: {self.start_line}\n"
            f"CU End line: {self.end_line}\n"
            f"Combinable with: {self.cwn_id}\n"
            f'in_dep: {" ".join(self.in_dep)}\n'
            f'out_dep: {" ".join(self.out_dep)}\n'
            f'in_out_dep: {" ".join(self.in_out_dep)}\n'
        )



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

import os
import shutil

import jsons  # type:ignore


class Settings(object):
    # general settings
    initialized = False
    discopop_build_dir: str = ""
    clang: str = ""
    clangpp: str = ""
    llvm_ar: str = ""
    llvm_link: str = ""
    llvm_dis: str = ""
    llvm_opt: str = ""
    llvm_llc: str = ""
    go_bin: str = ""
    use_docker_container_for_profiling: bool = True

    # code preview settings
    code_preview_show_metadata_regions: int = 0  # 1 = True, 0 = False
    code_preview_show_metadata_live_device_variables: int = 0  # 1 = True, 0 = False
    code_preview_show_line_numbers: int = 1  # 1 = True, 0 = False
    code_preview_disable_compile_check: int = 0  # 1 = True, 0 = False

    def __init__(self, discopop_build_dir="", go_bin_dir="", use_docker_container: bool = True) -> None:
        self.discopop_build_dir = discopop_build_dir
        self.go_bin = go_bin_dir
        self.use_docker_container_for_profiling = use_docker_container
        # validate settings
        settings_valid = os.path.exists(self.discopop_build_dir) and os.path.exists(self.go_bin)

        # get llvm_bin_dir from stored build configuration
        llvm_bin_dir = ""
        if settings_valid:
            command = "cat " + self.discopop_build_dir + '/build_config.txt | grep -oP "(?<=LLVM_BIN_DIR=).*"'
            llvm_bin_dir = os.popen(command).read().replace("\n", "")
            if not os.path.exists(llvm_bin_dir):
                llvm_bin_dir = ""

        # try and find default values using llvm_bin_dir and shutil.which
        if os.path.exists(os.path.join(llvm_bin_dir, "clang")):
            self.clang = os.path.join(llvm_bin_dir, "clang")
        else:
            self.clang = shutil.which("clang") or ""
        if os.path.exists(os.path.join(llvm_bin_dir, "clang++")):
            self.clangpp = os.path.join(llvm_bin_dir, "clang++")
        else:
            self.clangpp = shutil.which("clang++") or ""

        if os.path.exists(os.path.join(llvm_bin_dir, "llvm-ar")):
            self.llvm_ar = os.path.join(llvm_bin_dir, "llvm-ar")
        else:
            self.llvm_ar = shutil.which("llvm-ar-11") or ""

        if os.path.exists(os.path.join(llvm_bin_dir, "llvm-link")):
            self.llvm_link = os.path.join(llvm_bin_dir, "llvm-link")
        else:
            self.llvm_link = shutil.which("llvm-link-11") or ""

        if os.path.exists(os.path.join(llvm_bin_dir, "llvm-dis")):
            self.llvm_dis = os.path.join(llvm_bin_dir, "llvm-dis")
        else:
            self.llvm_dis = shutil.which("llvm-dis-11") or ""

        if os.path.exists(os.path.join(llvm_bin_dir, "opt")):
            self.llvm_opt = os.path.join(llvm_bin_dir, "opt")
        else:
            self.llvm_opt = shutil.which("opt-11") or ""

        if os.path.exists(os.path.join(llvm_bin_dir, "llc")):
            self.llvm_llc = os.path.join(llvm_bin_dir, "llc")
        else:
            self.llvm_llc = shutil.which("llc-11") or ""

        # validate settings
        settings_valid = (
            settings_valid
            and len(self.clang) > 0
            and len(self.clangpp) > 0
            and len(self.llvm_ar) > 0
            and len(self.llvm_link) > 0
            and len(self.llvm_dis) > 0
            and len(self.llvm_opt) > 0
            and len(self.llvm_llc) > 0
        )

        # set initialized, if all values could be determined and are valid, or docker container is used
        self.initialized = use_docker_container or settings_valid

    def get_as_json_string(self):
        """returns a representation of the settings which will be stored in a configuration file."""
        return jsons.dumps(self)

    def save_to_file(self, config_dir: str):
        settings_path = os.path.join(config_dir, "SETTINGS.txt")
        # remove old config if present
        if os.path.exists(settings_path):
            os.remove(settings_path)
        # write config to file
        with open(settings_path, "w+") as f:
            f.write(self.get_as_json_string())


def load_from_config_file(config_dir: str) -> Settings:
    json_str = ""
    with open(os.path.join(config_dir, "SETTINGS.txt"), "r") as f:
        for line in f.readlines():
            json_str += line
    value_dict = jsons.loads(json_str)
    settings = Settings()

    # general settings
    settings.discopop_build_dir = __load_or_get_default(value_dict, "discopop_build_dir")
    settings.clang = __load_or_get_default(value_dict, "clang")
    settings.clangpp = __load_or_get_default(value_dict, "clangpp")
    settings.llvm_ar = __load_or_get_default(value_dict, "llvm_ar")
    settings.llvm_link = __load_or_get_default(value_dict, "llvm_link")
    settings.llvm_dis = __load_or_get_default(value_dict, "llvm_dis")
    settings.llvm_opt = __load_or_get_default(value_dict, "llvm_opt")
    settings.llvm_llc = __load_or_get_default(value_dict, "llvm_llc")
    settings.go_bin = __load_or_get_default(value_dict, "go_bin")
    settings.use_docker_container_for_profiling = __load_or_get_default(
        value_dict, "use_docker_container_for_profiling"
    )
    # code preview settings
    settings.code_preview_show_metadata_regions = __load_or_get_default(
        value_dict, "code_preview_show_metadata_regions"
    )
    settings.code_preview_show_line_numbers = __load_or_get_default(value_dict, "code_preview_show_line_numbers")
    settings.code_preview_show_metadata_live_device_variables = __load_or_get_default(
        value_dict, "code_preview_show_metadata_live_device_variables"
    )
    settings.code_preview_disable_compile_check = __load_or_get_default(
        value_dict, "code_preview_disable_compile_check"
    )

    settings.initialized = True
    return settings


def __load_or_get_default(value_dict, key_name):
    if key_name in value_dict:
        return value_dict[key_name]
    # get default value
    return getattr(Settings(), key_name)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

"""Call clang++ with DiscoPoP LLVM passes."""

import argparse
import logging
import shutil

from discopop_library.global_data.version.utils import get_version
from . import DiscopopCpp

PROG = "discopop_profiler"

USAGE = f"""{PROG} [--verbose] [--clang CLANG]
       {'':{len(PROG)}} (--CUGeneration | --DPInstrumentation | --DPReduction)
       {'':{len(PROG)}} <clang++ arguments>
"""


def main(args=None):
    parser = argparse.ArgumentParser(prog=PROG, description=__doc__, usage=USAGE, add_help=False)
    parser.add_argument("-h", "--help", action="help", help="Show this help message and exit.")
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"%(prog)s {get_version()}",
        help="Show version number and exit.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Show additional information such as clang++ invocations.",
    )
    parser.add_argument("--clang", help="Path to clang++ executable.")
    action = parser.add_mutually_exclusive_group()
    action.add_argument(
        "--CUGeneration",
        "--cugeneration",
        action="store_true",
        help="Obtain the computational unit (CU) graph of the target application.",
    )
    action.add_argument(
        "--DPInstrumentation",
        "--dpinstrumentation",
        action="store_true",
        help="Instrument the target application to obtain data dependences.",
    )
    action.add_argument(
        "--DPReduction",
        "--dpreduction",
        action="store_true",
        help="Instrument the target application to obtain the list of reduction operations.",
    )
    parameters, clang_args = parser.parse_known_args(args)

    logging.basicConfig(
        format="%(message)s",
        level=logging.INFO if parameters.verbose else logging.WARNING,
    )

    if not any([parameters.CUGeneration, parameters.DPInstrumentation, parameters.DPReduction]):
        logging.warning(
            "Warning: Not using any DiscoPoP LLVM pass (specify either --CUGeneration, "
            "--DPInstrumentation or --DPReduction).",
        )
    clang_path = parameters.clang or shutil.which("clang++-8") or shutil.which("clang++")
    if not clang_path:
        raise SystemExit("clang++ executable not found in PATH. Specify --clang PATH/TO/CLANG++.")
    if not clang_args:
        logging.warning("Warning: No arguments to clang++ were given.")

    clang_proc = DiscopopCpp(
        cugeneration=parameters.CUGeneration,
        dpinstrumentation=parameters.DPInstrumentation,
        dpreduction=parameters.DPReduction,
        clang_path=clang_path,
    ).invoke(clang_args)
    if clang_proc.returncode != 0:
        raise SystemExit(clang_proc.returncode)


if __name__ == "__main__":
    main()



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from enum import IntEnum
from typing import Optional, List, Any


class PragmaPosition(IntEnum):
    BEFORE_START = 0
    AFTER_START = 1
    BEFORE_END = 2
    AFTER_END = 3


class Pragma(object):
    pragma_str: str = ""
    file_id: Optional[int] = None
    start_line: Optional[int] = None
    end_line: Optional[int] = None
    pragma_position: PragmaPosition = PragmaPosition.BEFORE_START
    parent_cu_id: str = ""
    children: List[Any] = []



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.

from typing import List, cast

from alive_progress import alive_bar  # type: ignore

from discopop_explorer.PEGraphX import PEGraphX, LoopNode
from discopop_explorer.pattern_detectors.PatternInfo import PatternInfo
from discopop_explorer.pattern_detectors.simple_gpu_patterns.GPULoop import GPULoopPattern
from discopop_explorer.pattern_detectors.simple_gpu_patterns.GPURegions import (
    GPURegions,
    GPURegionInfo,
)
from discopop_explorer.variable import Variable


def run_detection(pet: PEGraphX, res, project_folder_path: str) -> List[PatternInfo]:
    """Search for do-all loop pattern

    :param pet: PET graph
    :param res: DetectionResult object
    :return: List of detected pattern info
    """
    gpu_patterns: List[GPULoopPattern] = []

    loop_node_count = len(pet.all_nodes(type=LoopNode))
    print("\tcreate gpu patterns...")
    with alive_bar(loop_node_count) as progress_bar:
        for node in pet.all_nodes(type=LoopNode):
            # check for lastprivates, since they are not supported by the suggested pragma:
            #  pragma omp target teams distribute
            # todo: instead of omitting, suggest #pragma omp target parallel for instead
            if any(node.id == d.node_id for d in res.patterns.do_all if len(d.last_private) == 0) or any(
                node.id == r.node_id for r in res.patterns.reduction if len(r.last_private) == 0
            ):
                reduction_vars: List[Variable] = []
                if node.id in [r.node_id for r in res.patterns.reduction]:
                    parent_reduction = [r for r in res.patterns.reduction if r.node_id == node.id][0]
                    reduction_vars = parent_reduction.reduction
                gpulp = GPULoopPattern(
                    pet,
                    node.id,
                    node.start_line,
                    node.end_line,
                    node.loop_iterations,
                    project_folder_path,
                    reduction_vars,
                )
                gpulp.getNestedLoops(pet, node.id)
                gpulp.setParentLoop(node.id)
                gpulp.classifyLoopVars(pet, node)
                gpu_patterns.append(gpulp)
            progress_bar()

    if len(gpu_patterns) == 0:
        return []

    regions = GPURegions(pet, gpu_patterns, project_folder_path)

    for i in gpu_patterns:
        i.setCollapseClause(pet, i.node_id, res)

    regions.identifyGPURegions()
    # regions.old_mapData()
    regions.determineDataMapping()

    # construct GPURegions
    gpu_region_info: List[GPURegionInfo] = regions.get_gpu_region_info(pet, project_folder_path)

    return cast(List[PatternInfo], gpu_region_info)



# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import Dict, Set, List, Optional
import warnings

from sympy import Expr, Integer, Symbol  # type: ignore
import networkx as nx  # type: ignore

from discopop_explorer.PEGraphX import MemoryRegion
from discopop_library.discopop_optimizer.classes.context.Update import Update
from discopop_library.discopop_optimizer.classes.types.Aliases import DeviceID
from discopop_library.discopop_optimizer.classes.types.DataAccessType import (
    WriteDataAccess,
    ReadDataAccess,
)
from discopop_library.discopop_optimizer.utilities.simple_utilities import data_at


class ContextObject(object):
    last_visited_node_id: int
    last_visited_device_id: DeviceID
    last_seen_device_ids: List[DeviceID]
    seen_writes_by_device: Dict[DeviceID, Dict[MemoryRegion, Set[WriteDataAccess]]]
    necessary_updates: Set[Update]

    def __init__(self, initializing_node_id: int, last_seen_device_ids: Optional[List[DeviceID]] = None):
        self.seen_writes_by_device = dict()
        self.necessary_updates = set()
        self.last_visited_node_id = initializing_node_id
        self.last_visited_device_id = None
        self.last_seen_device_ids = last_seen_device_ids if last_seen_device_ids is not None else []
        self.snapshot_stack = []  # type: ignore
        self.save_stack = []  # type: ignore  # list of lists of ContextObjects, one list per branching depth

    def __str__(self):
        return str(self.necessary_updates)

    def calculate_and_perform_necessary_updates(
        self,
        node_reads: Set[ReadDataAccess],
        reading_device_id: int,
        reading_node_id: int,
        graph: nx.DiGraph,
        experiment,
    ):
        """checks if the specified list of ReadDataAccesses performed by the specified device id makes updates
        necessary. If so, the updates will get append to the list of updates of the current ContextObject.
        The list of seen writes by device of the ContextObject will be updated to reflect the identified data transfers.
        A reference to the object is returned."""
        required_updates: Set[Update] = set()
        for read in node_reads:
            # check if the reading device has the latest view of the memory
            for device_id in self.__get_known_device_ids():
                # todo test
                required_updates = set()

                if device_id == reading_device_id:
                    continue

                if read.memory_region not in self.get_seen_writes_by_device(device_id):
                    # read memory region is currently "unknown" to the device, thus is can be skipped
                    continue
                other_devices_known_writes = self.get_seen_writes_by_device(device_id)[read.memory_region]

                is_first_data_occurrence = False
                if read.memory_region not in self.get_seen_writes_by_device(reading_device_id):
                    # reading device does not currently "know" about the read memory region. create a new entry.
                    self.initialize_seen_writes_by_device(reading_device_id, read.memory_region)
                    is_first_data_occurrence = True

                known_writes = self.get_seen_writes_by_device(reading_device_id)[read.memory_region]
                unknown_writes = other_devices_known_writes.difference(known_writes)

                # todo debug: test: only consider the "latest" write
                unknown_writes_dict = dict()
                for entry in unknown_writes:
                    if entry.memory_region not in unknown_writes_dict:
                        unknown_writes_dict[entry.memory_region] = entry
                    else:
                        if unknown_writes_dict[entry.memory_region].unique_id < entry.unique_id:
                            # replace old write with "newer" write
                            unknown_writes_dict[entry.memory_region] = entry

                # unknown_writes = {max(unknown_writes, key=lambda x: x.unique_id)}
                unknown_writes = set(unknown_writes_dict.values())

                for data_write in unknown_writes:
                    # if device <-> device update is required, split it into two distinct updates
                    if (
                        device_id != experiment.get_system().get_host_device_id()
                        and reading_device_id != experiment.get_system().get_host_device_id()
                    ):
                        #                         print("Device <-> Device update required!")

                        # check if data is known to the host
                        if data_write.memory_region not in self.get_seen_writes_by_device(
                            experiment.get_system().get_host_device_id()
                        ):
                            self.initialize_seen_writes_by_device(
                                experiment.get_system().get_host_device_id(), data_write.memory_region
                            )
                        if (
                            data_write
                            not in self.get_seen_writes_by_device(experiment.get_system().get_host_device_id())[
                                data_write.memory_region
                            ]
                        ):
                            # register source device -> host update
                            required_updates.add(
                                Update(
                                    source_node_id=self.last_visited_node_id,
                                    target_node_id=reading_node_id,
                                    source_device_id=device_id,
                                    target_device_id=experiment.get_system().get_host_device_id(),  # reading_device_id,
                                    write_data_access=data_write,
                                    is_first_data_occurrence=is_first_data_occurrence,
                                    source_cu_id=data_at(graph, self.last_visited_node_id).original_cu_id,
                                    target_cu_id=data_at(graph, reading_node_id).original_cu_id,
                                )
                            )
                        #                        else:
                        #                            print(
                        #                                "SKIPPED KNOWN WRITE: ",
                        #                                str(
                        #                                    Update(
                        #                                        source_node_id=self.last_visited_node_id,
                        #                                        target_node_id=reading_node_id,
                        #                                        source_device_id=device_id,
                        #                                        target_device_id=0,  # reading_device_id,
                        #                                        write_data_access=data_write,
                        #                                        is_first_data_occurrence=is_first_data_occurrence,
                        #                                    )
                        #                                ),
                        #                            )

                        # register host -> target device update
                        required_updates.add(
                            Update(
                                source_node_id=self.last_visited_node_id,
                                target_node_id=reading_node_id,
                                source_device_id=0,  # device_id,
                                target_device_id=reading_device_id,
                                write_data_access=data_write,
                                is_first_data_occurrence=is_first_data_occurrence,
                                source_cu_id=data_at(graph, self.last_visited_node_id).original_cu_id,
                                target_cu_id=data_at(graph, reading_node_id).original_cu_id,
                            )
                        )

                    else:
                        # Host -> Device or Device -> Host update
                        required_updates.add(
                            Update(
                                source_node_id=self.last_visited_node_id,
                                target_node_id=reading_node_id,
                                source_device_id=device_id,
                                target_device_id=reading_device_id,
                                write_data_access=data_write,
                                is_first_data_occurrence=is_first_data_occurrence,
                                source_cu_id=data_at(graph, self.last_visited_node_id).original_cu_id,
                                target_cu_id=data_at(graph, reading_node_id).original_cu_id,
                            )
                        )

                # todo: check if this is sufficient
                for update in required_updates:
                    if update.write_data_access.memory_region not in self.get_seen_writes_by_device(
                        update.target_device_id
                    ):
                        self.initialize_seen_writes_by_device(
                            update.target_device_id, update.write_data_access.memory_region
                        )
                    self.__add_seen_write(
                        update.target_device_id, update.write_data_access.memory_region, update.write_data_access
                    )

                self.necessary_updates.update(required_updates)

        return self

    def add_writes(self, node_writes: Set[WriteDataAccess], writing_device_id: int):
        """Add the specified writes to the list of seen writes of the given device
        and returns a reference to this ContextObject."""
        # check if the device is known to the context
        if writing_device_id not in self.seen_writes_by_device:
            self.seen_writes_by_device[writing_device_id] = dict()

        for write in node_writes:
            # check if memory region is already present in self.seen_writes_by_device before adding the write access
            if write.memory_region not in self.get_seen_writes_by_device(writing_device_id):
                self.initialize_seen_writes_by_device(writing_device_id, write.memory_region)
            # add write to the list of seen writes
            self.__add_seen_write(writing_device_id, write.memory_region, write)
        return self

    def set_last_visited_node_id(self, node_id: int):
        self.last_visited_node_id = node_id

    def get_seen_writes_by_device(self, device_id: DeviceID) -> Dict[MemoryRegion, Set[WriteDataAccess]]:
        seen_dict: Dict[MemoryRegion, Set[WriteDataAccess]] = dict()

        # collect seen writes from stack
        for stack_entry in self.snapshot_stack:
            if device_id in stack_entry[0]:
                for memory_region in stack_entry[0][device_id]:
                    if memory_region not in seen_dict:
                        seen_dict[memory_region] = set()
                    seen_dict[memory_region].update(stack_entry[0][device_id][memory_region])

        # collect seen writes from self
        if device_id in self.seen_writes_by_device:
            for memory_region in self.seen_writes_by_device[device_id]:
                if memory_region not in seen_dict:
                    seen_dict[memory_region] = set()
                seen_dict[memory_region].update(self.seen_writes_by_device[device_id][memory_region])

        return seen_dict

    def initialize_seen_writes_by_device(self, device_id: DeviceID, memory_region: MemoryRegion):
        if device_id not in self.seen_writes_by_device:
            self.seen_writes_by_device[device_id] = dict()
        self.seen_writes_by_device[device_id][memory_region] = set()

    def __add_seen_write(self, device_id, memory_region: MemoryRegion, write: WriteDataAccess):
        if device_id not in self.seen_writes_by_device:
            self.seen_writes_by_device[device_id] = dict()
        if memory_region not in self.seen_writes_by_device[device_id]:
            self.seen_writes_by_device[device_id][memory_region] = set()
        self.seen_writes_by_device[device_id][memory_region].add(write)

    def __get_known_device_ids(self):
        seen_devices: Set[DeviceID] = set()
        for stack_entry in self.snapshot_stack:
            for device_id in stack_entry[0]:
                seen_devices.add(device_id)

        for device_id in self.seen_writes_by_device:
            seen_devices.add(device_id)

        return seen_devices
